/***********************************************************************
 ***  Type definiitions.
 ***
 ***  $Id: defs.h,v 1.1 2001/02/20 19:18:41 pberger Exp $
 ***********************************************************************/

#if sun
# define volatile
#endif

/*
 *  Unsigned types.
 */

#define vol volatile

typedef unsigned char uchar ;
typedef unsigned short ushort ;
typedef unsigned long ulong ;

/*
 *  The MIPS lint seems to have problems with casting functions
 *  returning a value as (void) and still produces a diagnostic.
 *  The define below avoids the problem.
 */

#if lint
int lintvoid ;
# define IGNORE lintvoid =
#else
# define IGNORE (void)
#endif

/*
 *  Structures used in more than one *.h file.   Defined here
 *  so order of *.h file definitions is easier.
 */

typedef struct con	CON ;
typedef struct host	HOST ;
typedef struct list	LIST ;
typedef	struct chan	CHAN ;
typedef	struct module	MOD ;
typedef struct pbus	PBUS ;

#if IUSC
typedef struct iusc	SC ;
#endif

#if SCA
typedef struct sca_ser	HS ;
typedef struct sca_chan	HC ;
typedef struct sca_dma	HD ;
#endif

typedef void LABEL() ;
typedef int FUNC() ;

/*
 *  Typedef for quantities in the channel structure
 *  whose volatility depends on whether there is a 
 *  host processor to look at them.
 */

#if SMART || SYNC
typedef volatile ushort hshort ;
typedef volatile uchar  hchar ;
#else
typedef ushort hshort ;
typedef uchar  hchar ;
#endif

/*
 *  Length of array macro.
 */

#define length(x)   (sizeof(x) / sizeof((x)[0]))

/*
 *  Macro to fail when something goes amiss.
 */

#if DBX
#  define crash			db_pause()
#else
#  define crash			panic(__FILE__, __LINE__)
#endif

#if lint || debug
#  define insist(x)		if (!(x)) crash
#  define assert(x)		if (!(x)) crash
#  define log(str, value)	logval(str, __LINE__, (ulong) value)
#else
#  define insist(x)		if (!(x)) crash
#  define assert(x)
#  define log(str, value)
#endif

#define REQUIRE(x) extern int _require_[x] ;

/*
 *  Macro for time comparisons.
 */

#define timege(x, y) ((ulong)(x) - (ulong)(y) < 0x40000000)

#define after(x) timege(stime, (x))


/****************************************************************
 ***	Defines to share structures between C and Assembly
 ****************************************************************/

#define MO(t,a)		(*(t *)(a))
#define VMO(t,a)	(*(volatile t *)(a))

#define	BEGIN_STRUCT(x)	struct x {
#define	S1(t,x)		t x ;
#define	S1A(t,x,n)	t x[n] ;
#define	S2(t,x)		t x ;
#define	S2A(t,x,n)	t x[n] ;
#define	S4(t,x)		t x ;
#define	S4A(t,x,n)	t x[n] ;
#define	END_STRUCT	} ;

#define ctrl(x)	((x) & 0x1f)


/***************************************************************
 ***  Bus error value.
 ***************************************************************/

#if lint
# define BUSERR	((ulong *)0x1)		/* Get bus error */
#else
# define BUSERR	((ulong *)0x1)		/* Get bus error */
#endif


/***************************************************************
 ***  Time base.
 ***************************************************************/

#define MS(t)	(t)		/* Time in milliseconds */
