/*
 * Copyright 1999,2000,2001 Digi International (www.digi.com)
 *    Chad Schwartz <chads@digi.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Based upon serial.c, written by Ted Ts'o.
 *  A big thank you goes to Ted, for a wonderful driver!
 *
 *
 *  Here are serial.c's release notes...
 *
 *  linux/drivers/char/serial.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  Extensively rewritten by Theodore Ts'o, 8/16/92 -- 9/14/92.  Now
 *  much more extensible to support other serial cards based on the
 *  16450/16550A UART's.  Added support for the AST FourPort and the
 *  Accent Async board.
 *
 *  set_serial_info fixed to set the flags, custom divisor, and uart
 * 	type fields.  Fix suggested by Michael K. Johnson 12/12/92.
 *
 *  11/95: TIOCMIWAIT, TIOCGICOUNT by Angelo Haritsis <ah@doc.ic.ac.uk>
 *
 *  03/96: Modularised by Angelo Haritsis <ah@doc.ic.ac.uk>
 *
 *  rs_set_termios fixed to look also for changes of the input
 *      flags INPCK, BRKINT, PARMRK, IGNPAR and IGNBRK.
 *                                            Bernd Anhupl 05/17/96.
 *
 *  1/97:  Extended dumb serial ports are a config option now.
 *         Saves 4k.   Michael A. Griffith <grif@acm.org>
 *
 *  8/97: Fix bug in rs_set_termios with RTS
 *        Stanislav V. Voronyi <stas@uanet.kharkov.ua>
 *
 *  3/98: Change the IRQ detection, use of probe_irq_o*(),
 *	  supress TIOCSERGWILD and TIOCSERSWILD
 *	  Etienne Lorrain <etienne.lorrain@ibm.net>
 *
 *  4/98: Added changes to support the ARM architecture proposed by
 * 	  Russell King
 *
 * This module exports the following rs232 io functions:
 *
 *	int rs_init(void);
 */

/***************************************************************************
March 17, 2000: CWS
-- Initial Release Version (This is the first version that
   actually has release notes - previous versions did not
   entirely work, therefore no release notes were created.)
-- Supports 4 and 8 port PCI ClassicBoards
-- Supports 32 ports across any combination of up to 4 boards
-- Uses dynamically allocated major numbers - grep /proc/devices
   for major number, and mknod. (There is a program that
   does this, that comes as a part of this driver's full
   package.)
-- Driver uses tq_immediate for its task queueing - No bottom
   half in the kernel is necessary.
-- Changed version to 1.0.0Beta1
June 27, 2000: CWS
-- Rid ourselves of MUCH extraneous code - such as the single port
   interrupt handler, all the leftover set_multiport stuff, etc.
-- Added a printd macro which just calls printf for you - IF DEBUG_CHAD1
   is defined.  If it isn't, the call doesn't do anything. (And is, in fact,
   100% optimized out of the code by the compiler.
-- Added a bunch of new UART types that were found in the new version of
   the Linux serial.c driver for Kernel 2.3.x.  This was done to maintain
   compliancy for prototype boards that hold MANY uarts, so we could test
   performance differences between them.
-- Went back to the compatibility mode for pci detection, to make kernel
   2.0 compliancy easier.
-- Added MMAP'ed I/O support to the driver. Now, one can dictate which
   mode they want for their particular board, based on a flag.
-- Previously, a LOT of places in the driver relied upon the fact that
   each port had an I/O address. Now, it can rely upon an I/O address OR'
   a memory address.
-- Corsair board now uses MMapped I/O instead of I/O mapped I/O
-- Added Multiple Vendor ID support to the driver - now, to add a new board
   that could possibly have a different VENDOR ID than Digi's vendor ID
   (0x114f) we can easily detect that board.
-- Added a "parity scan" function, which takes the place of the flip buffer
   scan that happens at the point that we fill the flip buffers.
-- Removed default support for FLIP buffers entirely.  We now just pass
   raw data off to tty->ldisc.receive_buf() once we have the data (in raw
   mode), or when NOT in raw mode, we spoof the flag buffer - so we can still
   check for parity errors, etc. with the new "parity scan" function.
-- Enhanced the various debug defines so they all follow 1 common format,
   rather than being virtually UNREADABLE
-- If a tty is throttled, DO NOT let autoflow make us start sending data
   again.
-- Bug in startup() that would clear all flow control states when
   the port is opened a second time.  This was solved by simply assuming
   that we do NOT need to run block_til_ready, if a port is completely open
   and running already.
-- Added size_fifo() function, which puts the UART in loopback mode,
   transmits 256 bytes, and then reads back bytes until the UART reports
   no more bytes to read.  This test seems to work, so I added it.  This code
   was grabbed from the serial.c in 2.3.x's version of serial.c
-- Added autoconfig_startech_uarts(), which allows us to do startech specific
   things, and free up all the autoconfig() mess that was beginning to grow.
   again, this code was taken from 2.3.x's version of serial.c
-- Got rid of register_serial() and unregister_serial() - I sincerely doubt
   we'll ever make a PCMCIA version of these boards. :)
-- Removed all references to detect_uart_irq() as we will *NEVER* use it.
-- Added support for the new prototype Exar 17C158 Octal UART/PCIset
   board.  We do support many of its extended features, such as Automatic
   hardware flow control, Memory Mapped FIFO's, 32 bit host <--> PCI
   transfers, and many other features.
-- Changed version to 1.1.1
January 29, 2001: Peter Berger (PEB)
-- Move the 4K "myflipbuf" from on stack in receive_chars() to
   cpci_async_struct in cpciP.h to fix the system degradation and lockups
   on 2.0.x kernels.
January 29, 2001: Al Borchers (ATB)
-- Defined MODVERSIONS before including modversions.h, as needed in 2.0
   kernels.
-- Changed to the new timer interface.  This is necessary for running on
   2.4 kernels.  Also, there is now no possible conflict with the built
   in serial driver.  There is still an unlikely module unload race in
   2.2 SMP kernels that has not been fixed; this was a problem with the
   old timer interface, too.
-- Changed kfree_s to kfree.  This is no longer supported in 2.4 kernels.
   Except for a consistency check kfree_s is equivalent to kfree.
-- Added initialization of wait queues for 2.4.0, and defined
   init_waitqueue_head for earlier kernels.
-- Used KERNEL_VERSION and defined it if needed.
February 18, 2001: PEB
-- Added DPA support
February 20, 2001: PEB/ATB; V1.1.4
-- Added support for new Digi Neo Vendor and PCI IDs
-- Fix device names displayed (to /var/log/messages) by cpci_init()
March 6, 2001: PEB; V1.1.5
-- Added 4-port Digi Neo support
March 12, 2001: PEB; V1.1.6
-- 4-port Acceleport XV fix.
-- Added DIGI_TIMERIRQ ioctl
-- Added DRVNAME to most printk calls
March 19, 2001: PEB; V1.1.7
-- Fix init_PCI() logic.
April 5, 2001: PEB; V1.1.8
-- Fixed kernel version dependent calls to request_irq() and free_irq()
April 18, 2001: PEB; V1.1.9
-- In cleanup_module() check if we are operating in I/O mapped mode before
   calling release_region() on the port.
April 23, 2001: PEB; V1.1.10
-- Comment out DIGI_TIMERIRQ printk message.
-- Added DIGI_LOOPBACK ioctl.
April 30, 2001: PEB: 1.1.11
-- Added support for new PCI interface routines when running newer kernels.
   Among other changes, newer kernels may remap the adapter's IRQ values
   and only the new interface routines will catch this.
May 4, 2001: PEB: 1.1.12
-- Fixed do-while loop logic in receive_chars()
May 29, 2001: PEB: 1.1.13
-- Added /proc support (/proc/tty/driver/cpci).
-- Set bd_type for Neo adapters.
May 31, 2001: PEB: 1.1.14
-- Converted C++ to C style comments.
-- Added ditty support.
June 7, 2001: PEB: 1.1.15
-- Added parm_prefix and parm_offset MODULE_PARM so the user diagnostics
   diskette can use non-Linux style names.  Note these are for informational
   messages (e.g: /proc) only.  Internally port counts start at 0 and
   serial_driver.name retains its value of "ttyCP" regardless of the
   parm settings.
June 9, 2001: PEB: 1.1.16
-- Fixed OLDPROC (for 2.0.X kernels) proc code
October 19, 2001: PEB: 1.1.16a
-- Fixed problem in cpci_interrupt() which caused hangs on Compaq ML530 server
   where RXDATA interrupts were being generated without setting the UART_LSR_DR
   bit so we never called receive_chars().  The workaround is to also check
   UART_EXTENDED_RFCNT.

???:???:v1.1.7-1
-- ???

April 3, 2003: PEB: v1.1.8-1
-- Added support for 2 port Neoclassic adapters (four versions:
	  2DB9           Neo 2p DB9 Univ PCI
	  2DB8PRI        Neo 2p DB9 Univ PCI  Powered Ring Indicator
	  2RJ45          Neo 2p RJ45 Univ PCI
	  2RJ45PRI       Neo 2p RJ45 Univ PCI  Powered Ring Indicator

April 18, 2003: PEB: V1.1.8-2
-- Added #ifdef'd RFCNT_KLUDGE code to workaround a debounce problem with
   the RFCNT register in the new XR17D15* UARTs.

April 24, 2003: PEB: V1.1.8-3
-- Changed RFCNT_KLUDGE code to do simply perform UART reads in single
   character mode.  Paradoxically this seems to give performance that
   matches or even slightly exceeds that of the deep FIFO (MEMMAPPED_MODE)
   code and incidentally also gets around the XR17D15* RFCNT register
   reporting problem.

April 28, 2003: PEB: V1.1.8-4
-- Fixed a buffer overflow condition activated by installing more than
   MAXBOARDS (currently set to 4) adapters.
-- Temporarily bumped MAXBOARDS to 5

May 15, 2003: SJD (Steve Dunlop) V1.1.8-5
-- Replaced RFCNT_KLUDGE with SINGLE_CHARACTER_READS

May 17, 2003: SJD: V1.1.8-8
-- Changed RX logic so that XOFF will not be sent more than once
-- Increased throughput performance by:
   1. Replacing the original wait-for-THRE-on-transmit logic with the
      new faster TX_TRIGGER_INTS code.  Both are workarounds for a
      problem in Rev C UARTs which causes them to occasionally report an
      incorrect TFCNT value.  The new approach provides a substantial
      performance boost.
   2. Replacing the original SINGLE_CHARACTER_READS code with the new
      RFCNT_WORKAROUND algorithm.  Both are workarounds for a problem
      in Rev D UARTs which causes them to occasionally report an incorrect
      RFCNT value.  The new approach provides a modest performance boost.
June 11, 2003: PEB: V1.1.8-10
-- Changed MAXBOARDS back to 4 for release and added -DDIAGS to set
   it to 8 when building the diagnostics diskettes.
June 15, 2003: PEB: V1.1.8-11
-- Added support for user-settable RX and TX FIFO triggers.

       

****************************************************************************/

/* These are CPCI specific defines */


#define TX_TRIGGER_INTS
#define RFCNT_WORKAROUND
#define TRIGGER_RX_17158 32
#define TRIGGER_TX_17158 10
// #define SINGLE_CHARACTER_READS

/*
 * TX_TRIGGER_INTS adds some transmit logic to deal with the TFCNT bug.
 * RFCNT_WORKAROUND uses memory mapped FIFO access except for the last
 * four bytes.
 */

#undef SERIAL_PARANOIA_CHECK

#define CONFIG_CPCI_NOPAUSE_IO
#define SERIAL_DO_RESTART

/* Normally these defines are controlled by the autoconf.h */

/* Set of debugging defines */

#define RS_STROBE_TIME (10*HZ)
#define RS_ISR_PASS_LIMIT 256

#define IRQ_T(state) SA_SHIRQ

#define SERIAL_INLINE

#if defined(MODULE) && defined(SERIAL_DEBUG_MCOUNT)
	#define DBG_CNT(s) printk("(%s): [%x] refc=%d, serc=%d, ttyc=%d -> %s\n", \
		kdevname(tty->device), (info->flags), serial_refcount,info->count,tty->count,s)
#else // !defined(MODULE) || !defined(SERIAL_DEBUG_MCOUNT)

	#define DBG_CNT(s)

#endif // !defined(MODULE) || !defined(SERIAL_DEBUG_MCOUNT)

#define DRVNAME serial_driver.name

#if defined DEBUG_CPCI

	#define DEBUG_PRINTK(string...) printk(string)

#else //! defined DEBUG

	#define DEBUG_PRINTK(string...)

#endif

#define DEBUG_PRINT_INTR(string...) DEBUG_PRINTK(string)
#define DEBUG_PRINT_TRANS(string...) DEBUG_PRINTK(string)
#define DEBUG_PRINT_RECV(string...) DEBUG_PRINTK(string)
#define DEBUG_PRINT_TIMER(string...) DEBUG_PRINTK(string)
/*
 * End of serial driver configuration section.
 */

#include <linux/version.h>
#include <linux/autoconf.h>

/* MODVERSIONS */
#ifdef CONFIG_MODVERSIONS
#       ifndef MODVERSIONS
#               define MODVERSIONS
#       endif
#       ifndef MODULE
#               define MODULE
#       endif
#	include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/proc_fs.h>
#include "serial.h"

/* GLOBAL 32BIT DEFINITIONS */

int cpci_totalboards;
int cpci_rawreadok = 1;
int cpci_use_rs_timer = 1;	/* Enable RS_TIMER interrupt use */

#include "rascon.h"
#include "dpacompat.h"
#include "pciclassic.h"
#include "cpciP.h"
#include "serial_reg.h"
#ifdef UTEST
#include UTESTDIR"utest.h"
#endif

#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)

	#include <linux/slab.h>

#else // LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)

	#include <linux/malloc.h>

#endif // LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)

#include <linux/delay.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/bitops.h>

#ifndef CONFIG_PCI
#error "You need to have PCI support enabled in your kernel for this driver"
#error "to function.  Enable PCI support in your kernel, and then reinstall"
#error "this driver package."
#endif

/* Need to use old /proc/fs routines on older kernels */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))

	#define OLDPROC

#endif // (LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0))


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)

	#include <linux/init.h>

#else // LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)

	#define mdelay(n) (\
		{unsigned long msec=(n); while (msec--) udelay(1000);})
	#define ioremap vremap
	#define iounmap vfree

#endif // LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)

/* Use the 2.4 timer interface with earlier kernels.  The semantics */
/* of del_timer_sync are not correct; del_timer_sync is supposed to */
/* wait for timers to finish.  This could be fixed, but it will */
/* only cause a problem in rare instances on module unload under 2.2 */
/* SMP kernels.  This was a problem using the old timer interface, */
/* too.  The mod_timer defined here should be correct; it is only used */
/* on 2.0 kernels, which are always UP kernels (if non-experimental), */
/* so the del_timer/add_timer will be atomic. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)

	/* Use old style PCI interface routines with old kernels. */
	#define OLDPCI 1

	#define del_timer_sync( t ) del_timer( t )

	#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)

		static inline int mod_timer( struct timer_list *timer, unsigned long expires )
		{
			int ret;
			unsigned long flags;

			DGLOCK(flags);
			timer->expires = expires;
			ret = del_timer( timer );
			add_timer( timer );
			DGUNLOCK(flags);
			return( ret );
		}

	#endif // LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)

#endif // LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)


/* init_waitqueue_head was introduced in 2.2.18 and 2.4.0 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)

	#ifndef init_waitqueue_head
		#define init_waitqueue_head(x) *(x)=NULL
	#endif

#endif // LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)

#include "serialasm.h"

#ifdef SERIAL_INLINE
#define _INLINE_ inline
#endif
	
/* Some compatibility stuff for 2.0.x */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)

	#include <linux/bios32.h>
	#include <linux/pci.h>
	#include <asm/segment.h>
	#define copy_from_user memcpy_fromfs
	#define copy_to_user memcpy_tofs
	#define putUser(arg1, arg2) put_fs_long(arg1, (unsigned long *)arg2)
	#define getUser(arg1, arg2) arg1 = get_fs_long((unsigned long *)arg2)
	/* compat declarations */
	extern int tty_get_baud_rate(struct tty_struct *tty);
	#define time_after(a,b)         ((long)(b) - (long)(a) < 0)
	#define time_before(a,b)        time_after(b,a)
	#define time_after_eq(a,b)      ((long)(a) - (long)(b) >= 0)
	#define time_before_eq(a,b)     time_after_eq(b,a)

#else // LINUX_VERSION_CODE > KERNEL_VERSION(2,1,0)

	#include <linux/init.h>
	#include <asm/uaccess.h>
	#include <linux/pci.h>
	#define putUser(arg1, arg2) put_user(arg1, (unsigned long *)arg2)
	#define getUser(arg1, arg2) get_user(arg1, (unsigned int *)arg2)

#endif // LINUX_VERSION_CODE > KERNEL_VERSION(2,1,0)

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,1,37))

	#define test_and_clear_bit clear_bit
	#define test_and_set_bit set_bit

#endif // (LINUX_VERSION_CODE < KERNEL_VERSION(2,1,37))

#if defined(MODULE_LICENSE)

	MODULE_LICENSE("GPL");

#endif

#if 0

	#include <linux/spinlock.h>
	spinlock_t cpci_lock = SPIN_LOCK_UNLOCKED;
	#define DGLOCK(flags) spin_lock_irqsave(&cpci_lock, flags)
	#define DGUNLOCK(flags) spin_unlock_irqrestore(&cpci_lock, flags)

#else

	#define DGLOCK(flags) do { save_flags(flags); cli( ); } while(0)
	#define DGUNLOCK(flags) do { restore_flags(flags); } while(0)

#endif

static char *cpci_name = "Digi International ClassicBoard PCI and Neo Driver";
static char *cpci_version = "v1.1.8-11";

static struct tty_driver serial_driver;
static struct tty_driver mgmt_driver;

static int serial_refcount;

struct timer_list cpci_timer_struct;

/* number of characters left in xmit buffer before we ask for more */
#define WAKEUP_CHARS 256

/*
 * IRQ_timeout		- How long the timeout should be for each IRQ
 * 				should be after the IRQ has been active.
 */

static struct cpci_async_struct *IRQ_ports[NR_IRQS];
static int IRQ_timeout[NR_IRQS];

static void autoconfig(struct cpci_serial_state * info);
static void change_speed(struct cpci_async_struct *info, struct termios *old);
static void cpci_wait_until_sent(struct tty_struct *tty, int timeout);
static void cpci_send_xchar(struct tty_struct *tty, char ch);

/*
 * Here we define the default xmit fifo size used for each type of
 * UART
 */
static struct serial_uart_config uart_config[] = {
{ "unknown", 1, 0 								},
{ "8250", 1, 0 									},
{ "16450", 1, 0 								},
{ "16550", 1, 0 								},
{ "16550A", 16, 	UART_CLEAR_FIFO | UART_USE_FIFO 			},
{ "cirrus", 1, 0 								},     /* usurped by cyclades.c */
{ "ST16650", 1, 	UART_CLEAR_FIFO | UART_STARTECH 			},
{ "ST16650V2", 32, 	UART_CLEAR_FIFO | UART_USE_FIFO	  | UART_STARTECH 	},
{ "TI16750", 64, 	UART_CLEAR_FIFO | UART_USE_FIFO				},
{ "Startech", 1, 0								},    /* usurped by cyclades.c */
{ "16C950/954", 128, 	UART_CLEAR_FIFO | UART_USE_FIFO				},
{ "ST16654", 64, 	UART_CLEAR_FIFO | UART_USE_FIFO   | UART_STARTECH 	},
{ "XR16C850", 128, 	UART_STARTECH 	| UART_CLEAR_FIFO | UART_USE_FIFO 	},
{ "XR17C158", 64, 	UART_MMAPED_FIFOS | UART_CLEAR_FIFO | UART_USE_FIFO   | UART_STARTECH 	},
/*{ "XR17C158(NOMMAP)", 64, UART_CLEAR_FIFO | UART_USE_FIFO | UART_STARTECH	},*/
{ 0, 0}
};

/* Module load time paramters */

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)

	/* Prefix and offset used for informational messages to identify a port */
	MODULE_PARM( parm_prefix, "s" );
	/* Port number 0 will be displayed using this offset */
	MODULE_PARM( parm_offset, "i" );

#endif // LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)

static char *parm_prefix="ttyCP";
static int parm_offset=0;


static struct cpci_serial_state cpci_table[MAXPORTS];

#define NR_PORTS	(sizeof(cpci_table)/sizeof(struct cpci_serial_state))

static struct tty_struct *serial_table[NR_PORTS];
static struct tty_struct *mgmt_table[NR_PORTS];

static struct termios *serial_termios[NR_PORTS];
static struct termios *serial_termios_locked[NR_PORTS];

/* This declaration is for a single-minor management device. it only */
/* needs 1 termios structure.						 */

static struct termios *mgmt_termios[1];
static struct termios *mgmt_termios_locked[1];

#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

/*
 * tmp_buf is used as a temporary buffer by serial_write.  We need to
 * lock it in case the copy_from_user blocks while swapping in a page,
 * and some other program tries to do a serial write at the same time.
 * Since the lock will only come under contention when the system is
 * swapping and available memory is low, it makes sense to share one
 * buffer across all the serial ports, since it significantly saves
 * memory if large numbers of serial ports are open.
 */
static unsigned char *tmp_buf;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)

	static DECLARE_MUTEX(tmp_buf_sem);

#else // LINUX_VERSION_CODE <= KERNEL_VERSION(2,3,0)

	static struct semaphore tmp_buf_sem = MUTEX;

#endif // LINUX_VERSION_CODE <= KERNEL_VERSION(2,3,0)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)

	static __inline__ int irq_cannonicalize(int irq)
	{
			return ((irq == 2) ? 9 : irq);
	}

#endif // LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,127)

	/* For older kernels that don't know about this function, lets redefine it */
	void schedule_timeout(int j) {
		current->state = TASK_INTERRUPTIBLE;
		current->timeout = jiffies + j;
		schedule();
		current->timeout = 0;
	}

#endif // LINUX_VERSION_CODE < KERNEL_VERSION(2,1,127)
				                
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
	
	/* 2.0.x kernels don't know anything about a tty_get_baud_rate() function */
	/* so lets declare it here.						  */
	static int baud_table[] = {
			0, 50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
			9600, 19200, 38400, 57600, 115200, 230400, 460800,
	#ifdef __sparc__
			76800, 153600, 307200, 614400, 921600
	#else
			500000, 576000, 921600, 1000000, 1152000, 1500000, 2000000,
			2500000, 3000000, 3500000, 4000000
	#endif
	};

	static int n_baud_table = sizeof(baud_table)/sizeof(int);

	int tty_get_baud_rate(struct tty_struct *tty)
	{
		
		unsigned int cflag, i;

		cflag = tty->termios->c_cflag;
		i = cflag & CBAUD;

		if (i & CBAUDEX) {
			i &= ~CBAUDEX;
			if (i < 1 || i+15 >= n_baud_table)
				tty->termios->c_cflag &= ~CBAUDEX;
			else
				i += 15;
		}
		
		return baud_table[i];
	}

#endif // LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)


static inline void serial_clear_cpci_table(void) {
	int i;
	
	for(i=0;i<NR_PORTS;i++) {
		memset(&cpci_table[i], 0, sizeof(struct cpci_serial_state));
		cpci_table[i].magic = 0;
		cpci_table[i].baud_base = BASE_BAUD;
		cpci_table[i].port = 0;
		cpci_table[i].irq = 3;
		cpci_table[i].flags = 0;
	}
}

static inline void ports_to_cpci_table(void) {
	int i;

	for(i=0;i<NR_PORTS;i++) {
		if(!ports[i].irq) {
			break;
		}
	
		cpci_table[i].port = ports[i].portaddr;
		cpci_table[i].irq = ports[i].irq;
		cpci_table[i].flags = STD_COM_FLAGS;
		cpci_table[i].xmit_fifo_size = ports[i].xmit_fifo_size;
		
		if(ports[i].io_type == CPCI_IO_MEM) {
			cpci_table[i].io_type =     CPCI_IO_MEM;
			cpci_table[i].iomem_base = ports[i].mem_portaddr;

		}		
		
		cpci_table[i].io_flags = ports[i].io_flags;
	}
}
	
int cpci_free_port_offset(void) {
	int i;

	for(i=0;i<MAXPORTS;i++){
		if(!ports[i].irq) return(i);
	}
	return(-1);	/* BAD - NO FREE PORTS	*/
}

static char *board_name( int bd_type )
{
	switch( bd_type ) {
		case DIGI_CLASSIC_4_PCI:
		case DIGI_CLASSIC_8_PCI:
			return( "ClassicBoard PCI" );
		case DIGI_NEOCLASSIC_4_PCI:
		case DIGI_NEOCLASSIC_8_PCI:

		case DIGI_NEOCLASSIC_2DB9_PCI:
		case DIGI_NEOCLASSIC_2DB9PRI_PCI:
		case DIGI_NEOCLASSIC_2RJ45_PCI:
		case DIGI_NEOCLASSIC_2RJ45PRI_PCI:
			return( "Neo" );
		default:
			return( "Unknown" );
	}

}

/* High water mark to avoid exceeding 4096 chars total */
#define PROC_HIWATER 4000

#ifdef OLDPROC

	int cpci_read_proc (char *page, char **start, off_t offset, int len, int unused)
	{
		int i, portcnt=0;

		len=0;
		len += sprintf( page,
				"%s %s.\n\n"
				"Found %d adapter%s:\n",
				cpci_name, cpci_version, cpci_totalboards,
				cpci_totalboards == 1? "" : "s" );

		for (i=0; i<cpci_totalboards && len < PROC_HIWATER; ++i) {
			portcnt += board[i].numports;
			len += sprintf( page+len,
				"  %d. Digi %s %d; IRQ:%2d; PCI(Bus:%d,Slot:%d); %s%d-%s%d\n",
				i+1, board_name(board[i].type), board[i].numports,
				board[i].irq,
				board[i].pci_bus, board[i].pci_slot,
				parm_prefix, board[i].firstchan+parm_offset,
				parm_prefix, board[i].firstchan+board[i].numports-1+parm_offset );
		}
		if( len < PROC_HIWATER )
			len += sprintf( page+len, "\n%d ports total.\n", portcnt );

		return len;
	}

	/* /proc/cpci info */
	struct proc_dir_entry cpci_proc_entry = {
		0,			/* low_ino: the inode */
		4, "cpci",		/* length and text of name */
		S_IFREG | S_IRUGO,	/* mode */
		1, 0, 0,		/* nlinks, owner, group */
		0,			/* size -- unused */
		NULL,			/* operations -- use default */
		&cpci_read_proc,	/* function used to read data */
	};

#else /* ! OLDPROC */

	int cpci_read_proc(char *page, char **start, off_t off, int count,
		int *eof, void *data)
	{
		int i, len=0, portcnt=0;
		off_t begin=0;

		*eof = 0;
		len += sprintf( page,
				"%s %s.\n\n"
				"Found %d adapter%s:\n",
				cpci_name, cpci_version, cpci_totalboards,
				cpci_totalboards == 1? "" : "s" );

		for (i=0; i<cpci_totalboards && len < PROC_HIWATER; ++i) {
			portcnt += board[i].numports;
			len += sprintf( page+len,
				"  %d. Digi %s %d; IRQ:%2d; PCI(Bus:%d,Slot:%d); %s%d-%s%d\n",
				i+1, board_name(board[i].type), board[i].numports,
				board[i].irq,
				board[i].pci_bus, board[i].pci_slot,
				parm_prefix, board[i].firstchan+parm_offset,
				parm_prefix, board[i].firstchan+board[i].numports-1+parm_offset );
			if (len+begin > off+count)
				goto done;

			if (len+begin < off) {
				begin += len;
				len = 0;
			}

		}
		if( len < PROC_HIWATER )
			len += sprintf( page+len, "\n%d ports total.\n", portcnt );
		*eof = 1;

	done:
		if (off >= len+begin)	/* Ran out of data before the request point */
			return 0;

		*start = page + (off-begin);
		return ((count < begin+len-off) ? count : begin+len-off);
	}

#endif // !OLDPROC

#ifdef OLDPCI

	typedef struct {
		unsigned vendor_id;
		unsigned pci_id;
		unsigned bd_type;
	} ID;

	ID PCI_IDs[] = {
		{DIGI_PCI_VENDOR_ID, DIGI_CLASSIC_4_PCI_ID, DIGI_CLASSIC_4_PCI},
		{DIGI_PCI_VENDOR_ID, DIGI_CLASSIC_8_PCI_ID, DIGI_CLASSIC_8_PCI},
		{EXAR_PCI_VENDOR_ID, DIGI_NEOCLASSIC_8_OLD_PCI_ID, DIGI_NEOCLASSIC_8_PCI},
		{DIGI_PCI_VENDOR_ID, DIGI_NEOCLASSIC_8_PCI_ID, DIGI_NEOCLASSIC_8_PCI},
		{DIGI_PCI_VENDOR_ID, DIGI_NEOCLASSIC_4_PCI_ID, DIGI_NEOCLASSIC_4_PCI},

		{DIGI_PCI_VENDOR_ID, DIGI_NEOCLASSIC_2DB9_PCI_ID, DIGI_NEOCLASSIC_2DB9_PCI},
		{DIGI_PCI_VENDOR_ID, DIGI_NEOCLASSIC_2DB9PRI_PCI_ID, DIGI_NEOCLASSIC_2DB9PRI_PCI},
		{DIGI_PCI_VENDOR_ID, DIGI_NEOCLASSIC_2RJ45_PCI_ID, DIGI_NEOCLASSIC_2RJ45_PCI},
		{DIGI_PCI_VENDOR_ID, DIGI_NEOCLASSIC_2RJ45PRI_PCI_ID, DIGI_NEOCLASSIC_2RJ45PRI_PCI},
	};

	#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
	int __init init_one_PCI (struct pci_dev *pdev, unsigned int bd_num)
	#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
	int init_one_PCI (struct pci_dev *pdev, unsigned int bd_num)
	#else
	__initfunc(int init_one_PCI (struct pci_dev *pdev, unsigned int bd_num) )
	#endif
	{
		board[bd_num].pci_bus = pdev->bus->number;
		board[bd_num].pci_slot = PCI_SLOT(pdev->devfn);

		switch(board[bd_num].type) {

			case DIGI_CLASSIC_4_PCI:
			case DIGI_CLASSIC_8_PCI:
				/* For PCI ClassicBoards
				 * PCI Local Address (i.e. "resource" number) space
				 * 0        PLX Memory Mapped Config
				 * 1        PLX I/O Mapped Config
				 * 2        I/O Mapped UARTs and Status
				 * 3        Memory Mapped VPD
				 * 4        Memory Mapped UARTs and Status
				 */
				board[bd_num].membase = pdev->base_address[4];
				board[bd_num].re_map_membase = ioremap((int)board[bd_num].membase, 0x100000);
				board[bd_num].iobase = pdev->base_address[2];

				/* Just an I/O offset - strip off bit 0 */
				board[bd_num].iobase = ((unsigned int)(board[bd_num].iobase)) & 0xFFFE;
				board[bd_num].pciset_iobase = pdev->base_address[1];
				/* Strip off bit 0... */
				board[bd_num].pciset_iobase = ((unsigned int)(board[bd_num].pciset_iobase)) & 0xFFFE;

				/* Grab board's IRQ. */
				board[bd_num].irq = pdev->irq;
				/* Grab board's IRQ Pin for Sharing Purposes */
				board[bd_num].irqpin = -1;
				break;
							
			case DIGI_NEOCLASSIC_4_PCI:
			case DIGI_NEOCLASSIC_8_PCI:
			case DIGI_NEOCLASSIC_2DB9_PCI:
			case DIGI_NEOCLASSIC_2DB9PRI_PCI:
			case DIGI_NEOCLASSIC_2RJ45_PCI:
			case DIGI_NEOCLASSIC_2RJ45PRI_PCI:
				/* This chip is set up 100% when we get to it. No need to	    	*/
				/* enable global interrupts or anything. Just have to  grab the IRQ 	*/			
				/* and membase, and map the memory for it...				*/

				board[bd_num].membase = pdev->base_address[0];
				
				board[bd_num].re_map_membase = ioremap((int)board[bd_num].membase, 0x1000);
				board[bd_num].irq = pdev->irq;
				board[bd_num].irqpin = -1;
				break;

			default:
				printk("%s: init_one_PCI() - Unknown card type\n",
					DRVNAME);
				return -EIO;
		}
		
		/* FLAG SET Switch */
		switch(board[bd_num].type) {
			case DIGI_CLASSIC_4_PCI:
				board[bd_num].numports = 4;
				board[bd_num].flags |= DIGI_PLX9052;
				break;
		
				
			case DIGI_CLASSIC_8_PCI:
				board[bd_num].numports = 8;
				board[bd_num].flags |= DIGI_PLX9052;
				break;
					
			case DIGI_NEOCLASSIC_4_PCI:
				board[bd_num].numports = 4;
				board[bd_num].flags |= DIGI_XR17C158;
				break;

			case DIGI_NEOCLASSIC_8_PCI:
				board[bd_num].numports = 8;
				board[bd_num].flags |= DIGI_XR17C158;
				break;

			case DIGI_NEOCLASSIC_2DB9_PCI:
			case DIGI_NEOCLASSIC_2DB9PRI_PCI:
			case DIGI_NEOCLASSIC_2RJ45_PCI:
			case DIGI_NEOCLASSIC_2RJ45PRI_PCI:
				board[bd_num].numports = 2;
				board[bd_num].flags |= DIGI_XR17C158;
				break;
					
			default:
				printk("%s: init_one_PCI() - Unknown card type\n",
					DRVNAME);
				return -EIO;
		}
			
  		/* NOTIFICATION and FINAL CONFIGURATION Switch */
		switch(board[bd_num].type) {
			case DIGI_CLASSIC_4_PCI:
			case DIGI_CLASSIC_8_PCI:		

				board[bd_num].dpatype = T_CLASSIC | T_PCIBUS;
		
				printk( "PCI Corsair %d Port (Card #%d;IRQ%d) Found\n",
					board[bd_num].numports,
					bd_num,
					board[bd_num].irq );
				/* Put all the proper information in the ports structure */
				/*cpci_plx9052_config_board(bd_num);*/
				board[bd_num].firstchan = cpci_free_port_offset();
				cpci_plx9052_config_board(bd_num, board[bd_num].firstchan);
				/* Enable global interrupt reporting for this card */
				cpci_plx9052_enable_interrupts(bd_num);
				break;
				
			
			case DIGI_NEOCLASSIC_4_PCI:
			case DIGI_NEOCLASSIC_8_PCI:
			case DIGI_NEOCLASSIC_2DB9_PCI:
			case DIGI_NEOCLASSIC_2DB9PRI_PCI:
			case DIGI_NEOCLASSIC_2RJ45_PCI:
			case DIGI_NEOCLASSIC_2RJ45PRI_PCI:

				board[bd_num].dpatype = T_NEO | T_PCIBUS;
				
				printk(
					"PCI Digi Neo %d Port (Card #%d;IRQ%d) Found\n",
					board[bd_num].numports,
					bd_num,
					board[bd_num].irq );
				/*cpci_xr17c158_config_board(bd_num);*/
				board[bd_num].firstchan = cpci_free_port_offset();
				cpci_xr17c158_config_board(bd_num, board[bd_num].firstchan);
				break;
			
			
			default:
				printk("%s: init_one_PCI() - Unknown card type\n",
					DRVNAME);
				return -EIO;
		}
	   	
		return(0);
	}

	#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
	int __init init_PCI(void)
	#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
	int init_PCI(void)
	#else
	__initfunc(int init_PCI(void))
	#endif
	{
		unsigned int boards_found = 0;
		int id;

		for (id = 0; id < (sizeof(PCI_IDs)/sizeof(ID)); id++)
		{
			struct pci_dev *pdev;
			pdev = NULL;
			
			while((pdev = pci_find_device(PCI_IDs[id].vendor_id,
							               PCI_IDs[id].pci_id, pdev)))
			{
				board[boards_found].type = PCI_IDs[id].bd_type;
				if( boards_found < MAXBOARDS ) {
					init_one_PCI(pdev, boards_found);
					boards_found++;
				} else {
					printk("%s: init_PCI() - MAXBOARDS(%d) exceeded; ignoring extra PCI_ID:%0xx board\n",
					DRVNAME, MAXBOARDS, PCI_IDs[id].pci_id );
				}
			}
		}

		return (boards_found);
	}

#else // !OLDPCI

	/* OLDPCI not defined, use new PCI interface */
	static struct pci_driver cpci_driver;

	static struct pci_device_id cpci_pci_tbl[] __initdata = {
		{ DIGI_PCI_VENDOR_ID, DIGI_CLASSIC_4_PCI_ID, PCI_ANY_ID, PCI_ANY_ID, 0, 0, DIGI_CLASSIC_4_PCI},
		{ DIGI_PCI_VENDOR_ID, DIGI_CLASSIC_8_PCI_ID, PCI_ANY_ID, PCI_ANY_ID, 0, 0, DIGI_CLASSIC_8_PCI},
		{ EXAR_PCI_VENDOR_ID, DIGI_NEOCLASSIC_8_OLD_PCI_ID, PCI_ANY_ID, PCI_ANY_ID, 0, 0, DIGI_NEOCLASSIC_8_PCI},
		{ DIGI_PCI_VENDOR_ID, DIGI_NEOCLASSIC_8_PCI_ID, PCI_ANY_ID, PCI_ANY_ID, 0, 0, DIGI_NEOCLASSIC_8_PCI},
		{ DIGI_PCI_VENDOR_ID, DIGI_NEOCLASSIC_4_PCI_ID, PCI_ANY_ID, PCI_ANY_ID, 0, 0, DIGI_NEOCLASSIC_4_PCI},

		{ DIGI_PCI_VENDOR_ID, DIGI_NEOCLASSIC_2DB9_PCI_ID, PCI_ANY_ID, PCI_ANY_ID, 0, 0, DIGI_NEOCLASSIC_2DB9_PCI},
		{ DIGI_PCI_VENDOR_ID, DIGI_NEOCLASSIC_2DB9PRI_PCI_ID, PCI_ANY_ID, PCI_ANY_ID, 0, 0, DIGI_NEOCLASSIC_2DB9PRI_PCI},
		{ DIGI_PCI_VENDOR_ID, DIGI_NEOCLASSIC_2RJ45_PCI_ID, PCI_ANY_ID, PCI_ANY_ID, 0, 0, DIGI_NEOCLASSIC_2RJ45_PCI},
		{ DIGI_PCI_VENDOR_ID, DIGI_NEOCLASSIC_2RJ45PRI_PCI_ID, PCI_ANY_ID, PCI_ANY_ID, 0, 0, DIGI_NEOCLASSIC_2RJ45PRI_PCI},
		{ 0, }
	};

	MODULE_DEVICE_TABLE(pci, cpci_pci_tbl);
	static unsigned int bd_num = 0;

	static int __init cpci_init_one (struct pci_dev *pdev,
				     const struct pci_device_id *ent)
	{
		unsigned int bd_type = ent->driver_data;

		if (pci_enable_device(pdev))
			return -EIO;

		if( bd_num >= MAXBOARDS ) {
			printk("%s: init_PCI() - MAXBOARDS(%d) exceeded; ignoring extra board at Bus:0x%x,Slot:0x%x\n",
			DRVNAME, MAXBOARDS, pdev->bus->number,PCI_SLOT(pdev->devfn) );
			return -EIO;
		}

		board[bd_num].pci_bus = pdev->bus->number;
		board[bd_num].pci_slot = PCI_SLOT(pdev->devfn);

		switch(bd_type) {

			case DIGI_CLASSIC_4_PCI:
			case DIGI_CLASSIC_8_PCI:
				/* For PCI ClassicBoards
				 * PCI Local Address (i.e. "resource" number) space
				 * 0        PLX Memory Mapped Config
				 * 1        PLX I/O Mapped Config
				 * 2        I/O Mapped UARTs and Status
				 * 3        Memory Mapped VPD
				 * 4        Memory Mapped UARTs and Status
				 */
				board[bd_num].membase = pdev->resource[4].start;
				board[bd_num].re_map_membase = ioremap((int)board[bd_num].membase, 0x100000);
				board[bd_num].type = bd_type;
				board[bd_num].iobase = pdev->resource[2].start;
				/* Just an I/O offset - strip off bit 0 */
				board[bd_num].iobase = ((unsigned int)(board[bd_num].iobase)) & 0xFFFE;
				board[bd_num].pciset_iobase = pdev->resource[1].start;
				/* Strip off bit 0... */
				board[bd_num].pciset_iobase = ((unsigned int)(board[bd_num].pciset_iobase)) & 0xFFFE;

				/* Grab board's IRQ. */
				board[bd_num].irq = pdev->irq;
				/* Grab board's IRQ Pin for Sharing Purposes */
				board[bd_num].irqpin = -1;
				break;
							
			case DIGI_NEOCLASSIC_4_PCI:
			case DIGI_NEOCLASSIC_8_PCI:
			case DIGI_NEOCLASSIC_2DB9_PCI:
			case DIGI_NEOCLASSIC_2DB9PRI_PCI:
			case DIGI_NEOCLASSIC_2RJ45_PCI:
			case DIGI_NEOCLASSIC_2RJ45PRI_PCI:
				/* This chip is set up 100% when we get to it. No need to	    	*/
				/* enable global interrupts or anything. Just have to  grab the IRQ 	*/			
				/* and membase, and map the memory for it...				*/

				board[bd_num].membase = pdev->resource[0].start;
				
				board[bd_num].re_map_membase = ioremap((int)board[bd_num].membase, 0x1000);
				board[bd_num].irq = pdev->irq;
				board[bd_num].irqpin = -1;
				board[bd_num].type = bd_type;
				break;

			default:
				printk("%s: init_one_PCI() - Unknown card type\n",
					DRVNAME);
				return -EIO;
		}
		
		/* FLAG SET Switch */
		switch(bd_type) {
			case DIGI_CLASSIC_4_PCI:
				board[bd_num].numports = 4;
				board[bd_num].flags |= DIGI_PLX9052;
				break;
		
				
			case DIGI_CLASSIC_8_PCI:
				board[bd_num].numports = 8;
				board[bd_num].flags |= DIGI_PLX9052;
				break;
					
			case DIGI_NEOCLASSIC_4_PCI:
				board[bd_num].numports = 4;
				board[bd_num].flags |= DIGI_XR17C158;
				break;

			case DIGI_NEOCLASSIC_8_PCI:
				board[bd_num].numports = 8;
				board[bd_num].flags |= DIGI_XR17C158;
				break;

			case DIGI_NEOCLASSIC_2DB9_PCI:
			case DIGI_NEOCLASSIC_2DB9PRI_PCI:
			case DIGI_NEOCLASSIC_2RJ45_PCI:
			case DIGI_NEOCLASSIC_2RJ45PRI_PCI:
				board[bd_num].numports = 2;
				board[bd_num].flags |= DIGI_XR17C158;
				break;
					
			default:
				printk("%s: init_one_PCI() - Unknown card type\n",
					DRVNAME);
				return -EIO;
		}
			
  		/* NOTIFICATION and FINAL CONFIGURATION Switch */
		switch(bd_type) {
			case DIGI_CLASSIC_4_PCI:
			case DIGI_CLASSIC_8_PCI:		

				board[bd_num].dpatype = T_CLASSIC | T_PCIBUS;
		
				printk( "PCI Corsair %d Port (Card #%d;IRQ%d) Found\n",
					board[bd_num].numports,
					bd_num,
					board[bd_num].irq );
				/* Put all the proper information in the ports structure */
				/*cpci_plx9052_config_board(bd_num);*/
				board[bd_num].firstchan = cpci_free_port_offset();
				cpci_plx9052_config_board(bd_num, board[bd_num].firstchan);
				/* Enable global interrupt reporting for this card */
				cpci_plx9052_enable_interrupts(bd_num);
				break;
				
			
			case DIGI_NEOCLASSIC_4_PCI:
			case DIGI_NEOCLASSIC_8_PCI:
			case DIGI_NEOCLASSIC_2DB9_PCI:
			case DIGI_NEOCLASSIC_2DB9PRI_PCI:
			case DIGI_NEOCLASSIC_2RJ45_PCI:
			case DIGI_NEOCLASSIC_2RJ45PRI_PCI:

				board[bd_num].dpatype = T_NEO | T_PCIBUS;
				
				printk(
					"PCI Digi Neo %d Port (Card #%d;IRQ%d) Found\n",
					board[bd_num].numports,
					bd_num,
					board[bd_num].irq );
				/*cpci_xr17c158_config_board(bd_num);*/
				board[bd_num].firstchan = cpci_free_port_offset();
				cpci_xr17c158_config_board(bd_num, board[bd_num].firstchan);
				break;
			
			
			default:
				printk("%s: init_one_PCI() - Unknown card type\n",
					DRVNAME);
				return -EIO;
		}
			
		bd_num++;
		return(0);
	}

	#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
	int __init init_PCI(void)
	#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
	int init_PCI(void)
	#else
	__initfunc(int init_PCI(void))
	#endif
	{
		int pci_count;

		memset (&cpci_driver, 0, sizeof (cpci_driver));
		cpci_driver.name = "cpci";
		cpci_driver.id_table = cpci_pci_tbl;
		cpci_driver.probe = cpci_init_one;

		bd_num = 0;	/* Incremented with each board found in cpci_init_one() */
		pci_count = pci_register_driver (&cpci_driver);

		if (pci_count <= 0) {
			pci_unregister_driver (&cpci_driver);
			pci_count = 0;
		}

		return(pci_count);
	}

#endif // !OLDPCI

void cpci_plx9052_enable_interrupts(int boardnum) {
	
	if(!board[boardnum].pciset_iobase) return;
	/* Send off the command to start up interrupts */
	
	outb(0x43, board[boardnum].pciset_iobase + 0x4c);
}

void cpci_plx9052_disable_interrupts(int boardnum) {
	/* If the board doesn't exist... */
	if(!board[boardnum].pciset_iobase) return;
	/* Send off the command to shut down interrupts.. */
	outb(0, board[boardnum].pciset_iobase + 0x4c);
}


int cpci_plx9052_config_board(int boardnum, int firstchan) {
	int i,offset=0;
	int freeport;
	struct board_struct *bd;
	
	bd = &board[boardnum];
	for(i=0;i<board[boardnum].numports;i++) {

		/*freeport = cpci_free_port_offset();*/
		freeport = firstchan+i;
		
		if(freeport == -1) {
			printk("%s: cpci_plx9052_config_board() - NO MORE FREE PORTS! BAD\n",
				DRVNAME);
			return(-1);
		}
		
		/* If for some reason you'd like to use the ClassicPCI board in */
		/* I/O mapped mode, uncomment the ports[freeport].portaddr 	*/
		/* assignment, and change the io_type to CPCI_IO_PORT - CWS	*/
		
		/*ports[freeport].portaddr = board[boardnum].iobase+offset;*/
	
		ports[freeport].portnum = freeport;	/* Index into ports[] */
		ports[freeport].cardnum = boardnum;
		ports[freeport].irq = board[boardnum].irq;
		
		ports[freeport].mem_portaddr = board[boardnum].re_map_membase + offset;
		ports[freeport].io_type = CPCI_IO_MEM;
		ports[freeport].io_flags = CPCI_IOFLAGS_REGULAR_EXTREGS;
		ports[freeport].xmit_fifo_size = board[boardnum].xmit_fifo_size;
		
		offset = offset+8;
	}
	
	return(0);
}
	
int cpci_xr17c158_config_board(int boardnum, int firstchan) {
	int i, offset=0;
	int freeport;
	struct board_struct *bd;
	
	bd = &board[boardnum];
	
	
	for(i=0;i<board[boardnum].numports;i++) {
		/*freeport = cpci_free_port_offset();*/
		freeport = firstchan+i;
		if(freeport == -1) {
			printk("%s: config_xr17c158_config_board() - NO MORE FREE PORTS! BAD\n", DRVNAME);
			return(-1);
		}
		
		ports[freeport].portnum = freeport;	/* Index into ports[] */
		ports[freeport].cardnum = boardnum;
		ports[freeport].irq = board[boardnum].irq;
		ports[freeport].mem_portaddr = board[boardnum].re_map_membase + offset;
		ports[freeport].io_type = CPCI_IO_MEM;
		ports[freeport].io_flags = CPCI_IOFLAGS_EXTENDED_EXTREGS;
		ports[freeport].xmit_fifo_size = board[boardnum].xmit_fifo_size;
		offset = offset+0x200;
	}

	return(0);
}

void cpci_xr17c158_set_debuglights(char *membase, char val) {
	/* ~val is done here because on the debug register setting */
	/* a bit means turning the light OFF, not ON. We want the  */
	/* function to APPEAR to the driver as though the exact    */
	/* opposite is true.					   */
	
	*(char*)(membase+XR17C158_PROTO_DEBUGLIGHTS) = (~val);
	udelay(400);
}
	

void cpci_plx9052_cleanup(int boardnum) {
	/* Disable global interrupts, and other stuff */
	cpci_plx9052_disable_interrupts(boardnum);
	
	/* Unmap base memory address */
	
	if(board[boardnum].re_map_membase) {
		iounmap(board[boardnum].re_map_membase);
		board[boardnum].re_map_membase = NULL;
	}
}

	
void cpci_xr17c158_cleanup(int boardnum) {
	/* Unmap base memory address */
	if(board[boardnum].re_map_membase) {
		iounmap(board[boardnum].re_map_membase);
		board[boardnum].re_map_membase = NULL;
	}
}
		

int init_boards(void) {
	printk("%s %s Loaded\n", cpci_name, cpci_version);
	printk("Copyright (C) 1999-2003 Digi International, Inc.\n");
	
	cpci_totalboards = init_PCI();	

	DEBUG_PRINTK("cpci_totalboards = %d\n",cpci_totalboards);
	
	if(!cpci_totalboards) {
		printk("%s: No boards found!\n", DRVNAME );
		return(-EBUSY);
	}
	return(0);
}

int cpci_internal_init(void) {
	ports_to_cpci_table();
	
	return(0);
}

/* Clean up interrupts and stuff, here. */

int cpci_internal_cleanup(void) {
	/* Lets gank the board's ability to generate interrupts */
	/* before we unload :) :)				*/
	int i=0;
	
	for(i=0;i<MAXBOARDS;i++) {
		if(board[i].flags & DIGI_PLX9052)
			cpci_plx9052_cleanup(i);
		else if(board[i].flags & DIGI_XR17C158)
			cpci_xr17c158_cleanup(i);
		}
		
	return(0);
}

static inline int serial_paranoia_check(struct cpci_async_struct *info,
					kdev_t device, const char *routine)
{
#ifdef SERIAL_PARANOIA_CHECK
	static const char *badmagic =
		"%s: Warning: bad magic number for serial struct (%s) in %s\n";
	static const char *badinfo =
		"%s: Warning: null cpci_async_struct for (%s) in %s\n";

	if (!info) {
		printk(badinfo, DRVNAME, kdevname(device), routine);
		return 1;
	}
	if (info->magic != CPCI_MAGIC) {
		printk(badmagic, DRVNAME, kdevname(device), routine);
		return 1;
	}
#endif
	return 0;
}

static inline unsigned int serial_in(struct cpci_async_struct *info, int offset)
{
	int val = 0;
	unsigned long flags;
	
	DGLOCK(flags);
	switch(info->io_type) {
		case CPCI_IO_MEM:
			val = readb(info->iomem_base + offset);
			break;
			
		default:
			val = inb(info->port + offset);
			break;
	}
	
	DGUNLOCK(flags);
	
	return(val);
}

static inline void serial_out(struct cpci_async_struct *info, int offset, int value)
{
 	unsigned long flags;
	
	DGLOCK(flags);
	switch(info->io_type) {
		case CPCI_IO_MEM:
			writeb(value, info->iomem_base + offset);
			break;
		
		default:
			outb(value, info->port+offset);
			break;
	}
	
	DGUNLOCK(flags);
}

/*
 * For the 16C950
 */

void serial_icr_write(struct cpci_async_struct *info, int offset, int  value)
{
	serial_out(info, UART_SCR, offset);
	serial_out(info, UART_ICR, value);
}
unsigned int serial_icr_read(struct cpci_async_struct *info, int offset)
{
	int     value;

	serial_icr_write(info, UART_ACR, info->ACR | UART_ACR_ICRRD);
	serial_out(info, UART_SCR, offset);
	value = serial_in(info, UART_ICR);
	serial_icr_write(info, UART_ACR, info->ACR);

	return value;
}


/*
 * ------------------------------------------------------------
 * cpci_stop() and cpci_start()
 *
 * This routines are called before setting or resetting tty->stopped.
 * They enable or disable transmitter interrupts, as necessary.
 * ------------------------------------------------------------
 */
static void cpci_stop(struct tty_struct *tty)
{
	struct cpci_async_struct *info = (struct cpci_async_struct *)tty->driver_data;
	unsigned long flags;

	if(tty->driver.subtype == SERIAL_TYPE_INFO)
		return;
		
	if (serial_paranoia_check(info, tty->device, "cpci_stop"))
		return;

	DGLOCK(flags);
	
	if (info->IER & UART_IER_THRI) {
		info->IER &= ~UART_IER_THRI;
		serial_out(info, UART_IER, info->IER);
	}
	
	if (info->state->type == PORT_16C950) {
		info->ACR |= UART_ACR_TXDIS;
		serial_icr_write(info, UART_ACR, info->ACR);
	}
	
	DGUNLOCK(flags);
}

static void cpci_start(struct tty_struct *tty)
{
	struct cpci_async_struct *info = (struct cpci_async_struct *)tty->driver_data;
	unsigned long flags;
	
	if(tty->driver.subtype == SERIAL_TYPE_INFO)
		return;

	if (serial_paranoia_check(info, tty->device, "cpci_start"))
		return;
	
	DGLOCK(flags);
	if (info->xmit_cnt && info->xmit_buf && !(info->IER & UART_IER_THRI)) {
		info->IER |= UART_IER_THRI;
		serial_out(info, UART_IER, info->IER);
	}

	if (info->state->type == PORT_16C950) {
		info->ACR &= ~UART_ACR_TXDIS;
		serial_icr_write(info, UART_ACR, info->ACR);
	}
	
	DGUNLOCK(flags);
}

static _INLINE_ void cpci_sched_event(struct cpci_async_struct *info,
				  int event)
{
	info->event |= 1 << event;

	/* - REMOVED - queue_task(&info->tqueue, &tq_cpci);*/
	/* - REMOVED - mark_bh(SERIAL_BH);*/
	
	queue_task(&info->tqueue, &tq_immediate);
	mark_bh(IMMEDIATE_BH);
}

#define CPCI_MEMMAPPED_MODE 0
#define CPCI_SINGLE_CHARACTER_MODE 1

static _INLINE_ void receive_chars(struct cpci_async_struct *info, int *status)
{
	struct tty_struct *tty = info->tty;
	int bytes_in_uart = 0;
	int number_bytes_to_copy = 0;
	int buffer_length = 0;
	int bytes_copied = 0;
	int copy_mode = CPCI_SINGLE_CHARACTER_MODE;
	char* buf_ptr = NULL;
	unsigned char ch = 0;
	
	if(cpci_rawreadok)
	{
		if(tty->real_raw)
		{
			buffer_length = sizeof(info->myflipbuf);
			buf_ptr = info->myflipbuf;
		}
		else// !tty->real_raw
		{
			buffer_length = sizeof(tty->flip.char_buf);
			buf_ptr = tty->flip.char_buf;
		}

		buffer_length = MIN(buffer_length, (N_TTY_BUF_SIZE - 1) - tty->read_cnt);
	}
	else// !cpci_rawreadok
	{
		buffer_length = TTY_FLIPBUF_SIZE - tty->flip.count;
		buf_ptr = tty->flip.char_buf_ptr;
	}

	DEBUG_PRINT_RECV("Entering receive_chars() on port %d.  Buffer = %d.\n",
				     MINOR(info->tty->device) + 1, buffer_length);

#ifndef SINGLE_CHARACTER_READS
/* If SINGLE_CHARACTER_READS is not defined then we can use the deep FIFOs, otherwise
 * we fall back on single character reads.
 *
 * The SINGLE_CHARACTER_READS are a workaround for the problem with the XR17D15
 * chip, which will, occasionally, report RFCNT incorrectly.  The memory-mapped
 * FIFOs don't really help with performance anyway.
 */
	if(uart_config[info->state->type].flags & UART_MMAPED_FIFOS)
	{
		DEBUG_PRINT_RECV("Setting memmapped mode on port %d.\n",
				         MINOR(info->tty->device) + 1);
		copy_mode = CPCI_MEMMAPPED_MODE;
	}
#endif

	do {
		/*
			If we detect an error somewhere in the 64 byte block,
			switch back to byte by byte mode.
		 */
		if(*status & UART_LSR_RECEIVE_FIFO_DATA_ERROR && !cpci_rawreadok)
		{
			DEBUG_PRINT_RECV("Error detected in port %d's uart, switching to single character mode.\n",
				             MINOR(info->tty->device) + 1);
			copy_mode = CPCI_SINGLE_CHARACTER_MODE;
		}

		if((buffer_length - bytes_copied) < 128)
		{
			DEBUG_PRINT_RECV("receive_chars(): Not enough space left in tty buffer = %d.  Disabling interrupts on port %d.\n",
				             (buffer_length - bytes_copied), MINOR(info->tty->device) + 1);
			if( info->IER & (UART_IER_RDI) ) {	/* sjd don't do this more than once or we can get */
				if (I_IXOFF(tty) )				/*     stuck sending XOFFs forever */
				{
					DEBUG_PRINT_RECV("Sending xoff character.\n");
					cpci_send_xchar(tty, STOP_CHAR(tty));
				}

				info->IER &= ~(UART_IER_RDI);
				serial_out(info, UART_IER, info->IER);
			}
		}

		switch(copy_mode)
		{
			case CPCI_MEMMAPPED_MODE:
				bytes_in_uart = serial_in(info, UART_EXTENDED_RFCNT);
#ifdef RFCNT_WORKAROUND
				bytes_in_uart -= 4;
#endif

			if( bytes_in_uart > 0 ) {
				number_bytes_to_copy = MIN(bytes_in_uart, (buffer_length - bytes_copied));
					if( number_bytes_to_copy > 0 )		//sjd safety
						memcpy((buf_ptr + bytes_copied), (info->iomem_base + UART_17158_RX_OFFSET), number_bytes_to_copy);

				bytes_copied += number_bytes_to_copy;
				DEBUG_PRINT_RECV("Copied %d bytes from port %d.\n",
						 bytes_copied, MINOR(info->tty->device) + 1);
			}
#ifdef RFCNT_WORKAROUND
		// FALL THROUGH to next case and get the last four bytes
#else
			break;
#endif
			case CPCI_SINGLE_CHARACTER_MODE:
			default:
				ch = serial_in(info, UART_RX);

				/*
				    Check if this character should be ignored, and
				    mask off conditions which should be ignored
				*/
				if(!(*status & info->ignore_status_mask))
				{
				    *(buf_ptr + bytes_copied) = ch;

					if(!cpci_rawreadok && !tty->real_raw)
					{
						if(*status & (UART_LSR_BI))
						{
				            DEBUG_PRINT_RECV("Set Break in flag buffer on port %d.\n",
				                             MINOR(info->tty->device) + 1);
				            *(tty->flip.flag_buf_ptr + bytes_copied) = TTY_BREAK;
						    info->state->icount.brk++;
						}
						else if(*status & UART_LSR_PE)
						{
							DEBUG_PRINT_RECV("Set Parity Error in flag buffer on port %d.\n",
				                             MINOR(info->tty->device) + 1);
				            *(tty->flip.flag_buf_ptr + bytes_copied) = TTY_PARITY;
							info->state->icount.parity++;
						}
						else if(*status & UART_LSR_FE)
						{
							DEBUG_PRINT_RECV("Set Frame Error in flag buffer on port %d.\n",
				                             MINOR(info->tty->device) + 1);
				            *(tty->flip.flag_buf_ptr + bytes_copied) = TTY_FRAME;
							info->state->icount.frame++;
						}
				        else
				        {
				            DEBUG_PRINT_RECV("Set Normal character in flag buffer on port %d.\n",
				                             MINOR(info->tty->device) + 1);
				            *(tty->flip.flag_buf_ptr + bytes_copied) = TTY_NORMAL;
				        }
					}
		
				    ++bytes_copied;
				    DEBUG_PRINT_RECV("Copied %d bytes from port %d.  Jiffies = %u.\n",
				                 bytes_copied, MINOR(info->tty->device) + 1, jiffies);
				}
				break;
		}
		
		*status = serial_in(info, UART_LSR);
	}
	while((*status & UART_LSR_DR) && (buffer_length - bytes_copied));
	
	info->state->icount.rx += bytes_copied;
	/*
		If we're using raw reads, pass it directly into the line disc,
		bypassing the flip buffers.
	*/
	if(cpci_rawreadok)
	{
		DEBUG_PRINT_RECV("Sending %d bytes directly to the line discipline for port %d.\n",
				          bytes_copied, MINOR(info->tty->device) + 1);
		if(tty->real_raw)
		{
			tty->ldisc.receive_buf(tty, info->myflipbuf, NULL, bytes_copied);
		}
		else// !tty->real_raw
		{
			tty->ldisc.receive_buf(tty, tty->flip.char_buf, tty->flip.flag_buf, bytes_copied);
		}
	}
	else// !cpci_rawreadok
	{
		DEBUG_PRINT_RECV("Scheduling flip for %d bytes on port %d.\n",
				          bytes_copied, MINOR(info->tty->device) + 1);
		tty->flip.count += bytes_copied;
		tty->flip.char_buf_ptr += bytes_copied;
	  	tty->flip.flag_buf_ptr += bytes_copied;
		tty_schedule_flip(tty);
	}
	
	DEBUG_PRINT_RECV("Leaving receive_chars() on port %d.\n", MINOR(info->tty->device) + 1);
}

static _INLINE_ void transmit_chars(struct cpci_async_struct *info)
{
	int count, bytesleft;
	int tmp;

	if (info->x_char) {
		serial_out(info, UART_TX, info->x_char);
		info->state->icount.tx++;
		info->x_char = 0;
	//	return;		// sjd not sure we should really return here, why not keep going?
	}

	if ( info->xmit_cnt <= 0 || info->tty->stopped || info->tty->hw_stopped){
		info->IER &= ~UART_IER_THRI;
		serial_out(info, UART_IER, info->IER);
		
		return;
	} 

	/*
	 * The XR17C* parts exhibitied a bug that would cause the EXTENDED_TFCNT
	 * register to report erroneously low values, resulting in transmit FIFO
	 * overfill and data loss.
	 *
	 * The original code used EXTENDED_TFCNT to calculate the exact number of
	 * empty bytes in the FIFO, and then filled it up all the way to the end.
	 * That didn't work properly, in some relatively rare cases, because of the
	 * hardware bug.  So a workaround was implemented based on a recommendation
	 * from Exar: Wait until THRE is set before filling the FIFO.  See the
	 * removed code in cpci_interrupt() where transmit_chars() is called.
	 *
	 * This caused transmit underruns, which adversely affected throughput.
	 *
	 * Our new workaround is to wait for THRI, which should not fire until
	 * the FIFO drains to the trigger level.  Then we fill the FIFO based
	 * on the minimum space that must be present when we are below trigger.
	 *
	 * As a slight optimization we check THRE and fill the fifo completely
	 * when it is set.
	 *
	 * sjd 5/03
	 */


	if(info->state->type == PORT_17158) {
#ifdef TX_TRIGGER_INTS
#ifndef TRIGGER_TX_17158
#error "you must define TRIGGER_TX_17158 if you are going to define TX_TRIGGER_INTS"
#endif
		if( serial_in( info, UART_LSR ) & UART_LSR_THRE )
			count = info->xmit_fifo_size;
		else
			count = info->xmit_fifo_size - TRIGGER_TX_17158;
#else
		// questionable workaround involving subtracting a few bytes from TFCNT
		// under the theory (not proven) that it's never off by very much
		count = info->xmit_fifo_size - serial_in( info, UART_EXTENDED_TFCNT );
		count -= 4;
		if( count <= 0 )
			return;
#endif
	} else {
		count = info->xmit_fifo_size;
	}
	
	if( !(uart_config[info->state->type].flags & UART_MMAPED_FIFOS)) {
	
		do {
			serial_out(info, UART_TX, info->xmit_buf[info->xmit_tail]);
			info->xmit_tail++;
			info->xmit_tail = info->xmit_tail & (CPCI_XMIT_SIZE-1);
			info->state->icount.tx++;
			if (--info->xmit_cnt <= 0)
				break;
		} while (--count > 0);
	} else {
		/* Memmapped FIFO's. */
		if(info->state->type != PORT_17158) {
			printk("%s: transmit_chars() - Port set for mmapped FIFOs, however UART\n", DRVNAME );
			printk("%s: transmit_chars() - is not an XR17C158! DISABLING UART!\n", DRVNAME );
			info->IER = 0;
			serial_out(info, UART_IER, info->IER);
			return;
		}
		
		/* For the mmaped I/O, we'll skip the xmit_cnt stuff we do */
		/* in the unmapped version, simply because we're NOT in a  */
		/* loop.  Lets just get the value as is, without including */
		/* it in some loop.                                        */
		
		count = MIN(count, info->xmit_cnt);
		
		/* We'll use this variable to store the bytes we have left */
		/* to copy from the ring buffer to a flat buffer           */

		bytesleft = count;

		do {
			if( ((info->xmit_tail + bytesleft) & (CPCI_XMIT_SIZE-1)) < info->xmit_tail) {
				/* Cache this value so we don't do the math multiple times */
				tmp = CPCI_XMIT_SIZE - info->xmit_tail;
				
				memcpy(info->iomem_base + UART_17158_TX_OFFSET,
					(info->xmit_buf + info->xmit_tail),
					tmp);
				
				info->xmit_cnt 	-= tmp;
				bytesleft	-= tmp;
				info->xmit_tail += tmp;
				
				info->xmit_tail &= (CPCI_XMIT_SIZE-1);		
			} else {
				memcpy(info->iomem_base + UART_17158_TX_OFFSET,
				(info->xmit_buf + info->xmit_tail),
				bytesleft);
				
				info->xmit_cnt -= bytesleft;
				info->xmit_tail += bytesleft;
				bytesleft = 0;
				info->xmit_tail &= (CPCI_XMIT_SIZE-1);
			}
		} while(bytesleft > 0);
	}
	
	if (info->xmit_cnt < WAKEUP_CHARS)
		cpci_sched_event(info, CPCI_EVENT_WRITE_WAKEUP);

	if (info->xmit_cnt <= 0) {
		info->IER &= ~UART_IER_THRI;
		serial_out(info, UART_IER, info->IER);
	}
}

static _INLINE_ void check_modem_status(struct cpci_async_struct *info)
{
	int	status;
	struct	cpci_async_icount *icount;
	status = serial_in(info, UART_MSR);
	
	if (status & UART_MSR_ANY_DELTA) {
		icount = &info->state->icount;
		/* update input line counters */
		if (status & UART_MSR_TERI)
			icount->rng++;
		if (status & UART_MSR_DDSR) {
			if( info->digiext.digi_flags & DIGI_ALTPIN )
				icount->dcd++;
			else
				icount->dsr++;
		}
		if (status & UART_MSR_DDCD) {
			if( info->digiext.digi_flags & DIGI_ALTPIN )
				icount->dsr++;
			else
				icount->dcd++;
		}
		if (status & UART_MSR_DCTS)
			icount->cts++;
		wake_up_interruptible(&info->delta_msr_wait);
	}

	/*if ((info->flags & ASYNC_CHECK_CD) && (status & UART_MSR_DDCD))*/
	if ((info->flags & ASYNC_CHECK_CD) &&
		(info->digiext.digi_flags & DIGI_ALTPIN?
		status & UART_MSR_DDSR :status & UART_MSR_DDCD))
	{
		DEBUG_PRINTK( "%s: ttys%d CD now %s..\n", DRVNAME, info->line,
			   (status & UART_MSR_DCD) ? "on" : "off");
		/*if (status & UART_MSR_DCD)*/
		if( info->digiext.digi_flags & DIGI_ALTPIN?
			status & UART_MSR_DSR: status & UART_MSR_DCD)
			wake_up_interruptible(&info->open_wait);
		else if (!((info->flags & ASYNC_CALLOUT_ACTIVE) &&
			   (info->flags & ASYNC_CALLOUT_NOHUP))) {
			DEBUG_PRINTK( "%s: doing serial hangup...\n", DRVNAME);
			if (info->tty)
				tty_hangup(info->tty);
		}
	}
	if (info->flags & ASYNC_CTS_FLOW) {
		if (info->tty->hw_stopped && !info->throttled) {
			if (status & UART_MSR_CTS) {
				DEBUG_PRINTK("%s: (minor %d) CTS tx start...\n", DRVNAME, MINOR(info->tty->device));
				info->tty->hw_stopped = 0;
				info->IER |= UART_IER_THRI;
				serial_out(info, UART_IER, info->IER);
				cpci_sched_event(info, CPCI_EVENT_WRITE_WAKEUP);
				return;
			}
		} else {
			if (!(status & UART_MSR_CTS)) {
				DEBUG_PRINTK("%s: (minor %d) CTS tx stop...\n", DRVNAME, MINOR(info->tty->device));
				info->tty->hw_stopped = 1;
				info->IER &= ~UART_IER_THRI;
				serial_out(info, UART_IER, info->IER);
			}
		}
	}
}

/*
 * This is the serial driver's generic interrupt routine
 */
static void cpci_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	int line_status_register, interrupt_status_register;
	struct cpci_async_struct *info;
	int  once_kludge = 1;
	
	DEBUG_PRINT_INTR("%s: Entering cpci_interrupt(%d)...\n", DRVNAME, irq);

	info = IRQ_ports[irq];
	
	if (info) {
		do {
			once_kludge = 1;
			if( info->tty ) {
#ifdef TX_TRIGGER_INTS
				while ( !((interrupt_status_register = serial_in(info, UART_ISR)) & UART_ISR_NO_INT) || once_kludge )
#else
				interrupt_status_register = serial_in(info, UART_ISR);
				/* must not read this more than once as some bits are self clearing */
#endif
				{

					DEBUG_PRINT_INTR("\nEntering interrupt routine for port %d.\n",
									 (MINOR(info->tty->device) + 1));
					info->last_active = jiffies;
			
					if(!(info->IER & UART_IER_RDI) &&
					   (((N_TTY_BUF_SIZE - 1) - info->tty->read_cnt) > (N_TTY_BUF_SIZE / 2))) {
						/*
							If we have disabled receive interrupts due to a full tty
							buffer, re-enable when we have space.
						*/
						DEBUG_PRINT_INTR("\nEnabling receive interrupts on port %d.\n",
										 (MINOR(info->tty->device) + 1));
						info->IER |= UART_IER_RDI;
						serial_out(info, UART_IER, info->IER);
			
						if(!info->throttled && I_IXOFF(info->tty)) {	// SJD don't XON if throttled
							cpci_send_xchar(info->tty, START_CHAR(info->tty));
						}
					}
			
					line_status_register = serial_in(info, UART_LSR);
/* don't need to do this twice
					if(!(uart_config[info->state->type].flags & UART_STARTECH) &&
					   !info->loopedback && (interrupt_status_register & 0x20)) {
						check_modem_status(info);
					}
*/
					
					if(info->state->type == PORT_17158) {
					    if( once_kludge ) {
							if (line_status_register & UART_LSR_DR || serial_in(info, UART_EXTENDED_RFCNT)) 
								receive_chars(info, &line_status_register);
					    } else {
							if( (interrupt_status_register & UART_ISR_RDI) ||
									(interrupt_status_register&0x20) ||
									(serial_in(info,UART_EXTENDED_RFCNT)>32) )
								receive_chars(info, &line_status_register);
					    }
					} else {
						if (line_status_register & UART_LSR_DR) {
							receive_chars(info, &line_status_register);
						}
					}
			
					if(line_status_register & UART_LSR_OE &&
					   !(line_status_register & info->ignore_status_mask)) {
						printk("%s: FIFO OVERRUN on port %d!\n", DRVNAME,
							   MINOR(info->tty->device) + 1);
					}
					
					check_modem_status(info);
					
#ifdef TX_TRIGGER_INTS
					if( info->state->type != PORT_17158 ||
							(interrupt_status_register & UART_ISR_THRI) ||
							(line_status_register & UART_LSR_THRE) )
#endif
					{
						transmit_chars(info);
					}
					
					DEBUG_PRINT_INTR("Leaving interrupt routine for port %d.\n\n",
									 (MINOR(info->tty->device) + 1));
					once_kludge = 0;
				}
			}
			
			info = info->next_port;
		} while (info);	

	} else {
		printk("%s: cpci_interrupt(%d) - No info structure! Dying!\n",
			   DRVNAME, irq);
	}

	DEBUG_PRINT_INTR("%s: Leaving cpci_interrupt()\n", DRVNAME);
}

/*
 * -------------------------------------------------------------------
 * Here ends the serial interrupt routines.
 * -------------------------------------------------------------------
 */

static void do_softint(void *private_)
{
	struct cpci_async_struct	*info = (struct cpci_async_struct *) private_;
	struct tty_struct	*tty;
	unsigned long flags;
	
	tty = info->tty;
	if (!tty)
		return;

	DGLOCK(flags);
	if (test_and_clear_bit(CPCI_EVENT_WRITE_WAKEUP, &info->event)) {
		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
			tty->ldisc.write_wakeup)
			(tty->ldisc.write_wakeup)(tty);
		wake_up_interruptible(&tty->write_wait);
	}
	
	DGUNLOCK(flags);
}

/*
 * This subroutine is called when the RS_TIMER goes off.  It is used
 * by the serial driver to handle ports that do not have an interrupt
 * (irq=0).  This doesn't work very well for 16450's, but gives barely
 * passable results for a 16550A.  (Although at the expense of much
 * CPU overhead).
 */
static void cpci_timer( unsigned long unused )
{
	static unsigned long last_strobe = 0;
	struct cpci_async_struct *info;
	unsigned int	i;
	unsigned long flags;

	/* Ignore timer interrupts: Using adapter interrupts only */
	if( !cpci_use_rs_timer )
	{
		mod_timer( &cpci_timer_struct, jiffies + RS_STROBE_TIME );
		return;
	}

	DEBUG_PRINT_TIMER("Entering cpci_timer().\n");

	if ((jiffies - last_strobe) >= RS_STROBE_TIME)
	{
		for (i=1; i < NR_IRQS; i++)
		{
			info = IRQ_ports[i];
			
			if (!info)
			{
				continue;
			}
			
			DGLOCK(flags);

			do {
				serial_out(info, UART_IER, 0);
				info->IER |= UART_IER_THRI;
				serial_out(info, UART_IER, info->IER);
				info = info->next_port;
			} while (info);

			cpci_interrupt(i, NULL, NULL);

			DGUNLOCK(flags);
		}
	}
	
	last_strobe = jiffies;
	mod_timer( &cpci_timer_struct, jiffies + RS_STROBE_TIME );

	if (IRQ_ports[0])
	{
		DGLOCK(flags);
		cpci_interrupt(0, NULL, NULL);
		DGUNLOCK(flags);

		mod_timer( &cpci_timer_struct, jiffies + IRQ_timeout[0] - 2 );
	}

	DEBUG_PRINT_TIMER("Leaving cpci_timer().\n");
}

/*
 * This routine figures out the correct timeout for a particular IRQ.
 * It uses the smallest timeout of all of the serial ports in a
 * particular interrupt chain.  Now only used for IRQ 0....
 */

static void figure_IRQ_timeout(int irq)
{
	struct	cpci_async_struct	*info;
	int	timeout = 60*HZ;	/* 60 seconds === a long time :-) */

	info = IRQ_ports[irq];
	if (!info) {
		IRQ_timeout[irq] = 60*HZ;
		return;
	}
	while (info) {
		if (info->timeout < timeout)
			timeout = info->timeout;
		info = info->next_port;
	}
	if (!irq)
		timeout = timeout / 2;
	IRQ_timeout[irq] = timeout ? timeout : 1;
}

static int startup(struct cpci_async_struct * info)
{
	unsigned long flags;
	int	retval=0;
	void (*handler)(int, void *, struct pt_regs *);
	struct cpci_serial_state *state= info->state;
	unsigned long page;
	
	if(info->state->count > 1)
		return(0);
			
	page = get_free_page(GFP_KERNEL);
	if (!page)
		return -ENOMEM;

	DGLOCK(flags);

	if (info->flags & ASYNC_INITIALIZED) {
		free_page(page);
		goto errout;
	}

	if(!CONFIGURED_SERIAL_PORT(info)) {
		if (info->tty)
			set_bit(TTY_IO_ERROR, &info->tty->flags);
		free_page(page);
		goto errout;
	}
	if (info->xmit_buf)
		free_page(page);
	else
		info->xmit_buf = (unsigned char *) page;

	DEBUG_PRINTK("%s: starting up ttys%d (irq %d)...\n", DRVNAME, info->line, state->irq);

	if (uart_config[info->state->type].flags & UART_STARTECH) {
		/* Wake up UART */
		
		/* If it isn't an extended flags UART (Aka a 17C158...) */

		if(info->state->type != PORT_17158) {
			serial_out(info, UART_LCR, 0xBF);
			serial_out(info, UART_EFR, UART_EFR_ECB);
			serial_out(info, UART_IER, 0);
			serial_out(info, UART_EFR, 0);
		} else {
			/* Gotta be some a 17158 startech uart... */
			
			serial_out(info, UART_EXTENDED_EFR, UART_EFR_ECB);
			serial_out(info, UART_IER, 0);
		}
		
		/** For a XR16C850, we need to set the trigger levels*/
		if (info->state->type == PORT_16850) {
			serial_out(info, UART_FCTR, UART_FCTR_TRGD | UART_FCTR_RX);
			serial_out(info, UART_TRG, UART_TRG_96);
			serial_out(info, UART_FCTR, UART_FCTR_TRGD | UART_FCTR_TX);
			serial_out(info, UART_TRG, UART_TRG_96);
		}
		
		if (info->state->type == PORT_17158) {

			/* Set the trigger level for 32 bytes, with a Hysteresis */
			/* level of 8. That should allow for an additional 16    */
			/* bytes to land in the FIFO, if a remote 16550 throws 	 */
			/* its entire TX fifo at us - while being unable to stop */
			/* on a dime, with hardware flow control		 */

			/* Note: rxtrg and txtrg are now set in get_cpci_async_struct()
			 * on the first open only so subsequent user settings will
			 * remain sticky */
			/* info->digiext.rxtrg = 32;*/
			serial_out(info, UART_EXTENDED_RXTRG, info->digiext.rxtrg);
			serial_out(info, UART_EXTENDED_FCTR, UART_17158_FCTR_TRGD | UART_17158_FCTR_RTS_8DELAY);

			// sjd, trig for tx int
#ifdef TRIGGER_TX_17158
			/*info->digiext.txtrg = TRIGGER_TX_17158;*/
			serial_out( info, UART_EXTENDED_TXTRG, info->digiext.txtrg );
#endif
		}

		serial_out(info, UART_LCR, 0);
	}

	if (info->state->type == PORT_16750) {
		/* Wake up UART */
		serial_out(info, UART_IER, 0);
	}

	if (info->state->type == PORT_16C950) {
		/* Wake up and initialize UART */
		info->ACR = 0;
		serial_out(info, UART_LCR, 0xBF);
		serial_out(info, UART_EFR, UART_EFR_ECB);
		serial_out(info, UART_IER, 0);
		serial_out(info, UART_LCR, 0);
		serial_icr_write(info, UART_CSR, 0); /* Reset the UART */
		serial_out(info, UART_LCR, 0xBF);
		serial_out(info, UART_EFR, UART_EFR_ECB);
		serial_out(info, UART_LCR, 0);
	}
	
	if (info->state->type == PORT_17158) {
		/* Wake up and initialize UART */
		
		serial_out(info, UART_EXTENDED_EFR, UART_EFR_ECB);
		serial_out(info, UART_IER, 0);
		serial_out(info, UART_LCR, 0);
	}
	
	/*
	 * Clear the FIFO buffers and disable them
	 * (they will be reenabled in change_speed())
	 */
	if (uart_config[state->type].flags & UART_CLEAR_FIFO) {
	
		serial_out(info, UART_FCR, UART_FCR_ENABLE_FIFO);
		serial_out(info, UART_FCR, (UART_FCR_ENABLE_FIFO |
						UART_FCR_CLEAR_RCVR |
						UART_FCR_CLEAR_XMIT));
		serial_out(info, UART_FCR, 0);
	}
	
	/*
	 * Clear the interrupt registers.
	 */
	
	(void) serial_in(info, UART_LSR);
	(void) serial_in(info, UART_RX);
	(void) serial_in(info, UART_ISR);
	
	/*
	 * At this point there's no way the LSR could still be 0xFF;
	 * if it is, then bail out, because there's likely no UART
	 * here.
	 */
	if (serial_in(info, UART_LSR) == 0xff) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
		if(suser())
#else
		if (capable(CAP_SYS_ADMIN))
#endif
		{
			if (info->tty)
				set_bit(TTY_IO_ERROR, &info->tty->flags);
		} else
			retval = -ENODEV;
		goto errout;
	}
	
	/*
	 * Allocate the IRQ if necessary
	 */
	if (state->irq && (!IRQ_ports[state->irq] ||
			  !IRQ_ports[state->irq]->next_port)) {

		if (IRQ_ports[state->irq]) {
			free_irq(state->irq, &IRQ_ports[state->irq]);
		}
		
		handler = cpci_interrupt;
		retval = request_irq(state->irq, handler, SA_SHIRQ,
					 "ClassicPCI", &IRQ_ports[state->irq]);

		if (retval) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
			if (suser())
#else
			if (capable(CAP_SYS_ADMIN))
#endif
			{
				if (info->tty)
					set_bit(TTY_IO_ERROR,
						&info->tty->flags);
				retval = 0;
			}
			goto errout;
		}
	}

	/*
	 * Insert serial port into IRQ chain.
	 */
	info->prev_port = 0;
	info->next_port = IRQ_ports[state->irq];
	if (info->next_port)
		info->next_port->prev_port = info;
	IRQ_ports[state->irq] = info;
	figure_IRQ_timeout(state->irq);

	/*
	 * Clear the interrupt registers.
	 */
	 /* (void) serial_in(info, UART_LSR); */   /* (see above) */
	(void) serial_in(info, UART_RX);
	(void) serial_in(info, UART_ISR);
	(void) serial_in(info, UART_MSR);

	/*
	 * Now, initialize the UART
	 */
	serial_out(info, UART_LCR, UART_LCR_WLEN8);	/* reset DLAB */

	info->MCR = 0;
	if (info->tty->termios->c_cflag & CBAUD)
		info->MCR = UART_MCR_DTR | UART_MCR_RTS;
		if (state->irq != 0) {
			if(info->state->type == PORT_17158) {
			} else {
				info->MCR |= UART_MCR_OUT2;
			}
		}
	 	
	 	if(info->loopedback) {
	 		printk("%s: Enabling UART loopback on line:%d\n", DRVNAME, info->line );
	 		info->MCR |= UART_MCR_LOOP;
	 	}
	 	
		serial_out(info, UART_MCR, info->MCR);
	
	/*
	 * Finally, enable interrupts
	 */
	info->IER = UART_IER_MSI | UART_IER_RLSI | UART_IER_RDI;
	
	if(info->state->type == PORT_17158)
		info->IER |= UART_17158_IER_RTSDTR | UART_17158_IER_CTSDSR;
	
	serial_out(info, UART_IER, info->IER);	/* enable interrupts */
	

	/*
	 * And clear the interrupt registers again for luck.
	 */
	(void)serial_in(info, UART_LSR);
	(void)serial_in(info, UART_RX);
	(void)serial_in(info, UART_ISR);
	(void)serial_in(info, UART_MSR);

	if (info->tty)
		clear_bit(TTY_IO_ERROR, &info->tty->flags);
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;

	/*
	 * Set up serial timers...
	 */
	mod_timer( &cpci_timer_struct, jiffies + 2*HZ/100 );

	/*
	 * Set up the tty->alt_speed kludge
	 */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	if (info->tty) {
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI)
			info->tty->alt_speed = 57600;
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI)
			info->tty->alt_speed = 115200;
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
			info->tty->alt_speed = 230400;
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
			info->tty->alt_speed = 460800;
	}
#endif	
	/*
	 * and set the speed of the serial port
	 */
	change_speed(info, 0);

	info->flags |= ASYNC_INITIALIZED;
	DGUNLOCK(flags);
	return 0;
	
errout:
	DGUNLOCK(flags);
	return retval;
}

/*
 * This routine will shutdown a serial port; interrupts are disabled, and
 * DTR is dropped if the hangup on close termio flag is on.
 */
static void shutdown(struct cpci_async_struct * info)
{
	unsigned long	flags;
	struct cpci_serial_state *state;
	int		retval;
	
	if (!(info->flags & ASYNC_INITIALIZED))
		return;

	state = info->state;

	DEBUG_PRINTK("%s: Shutting down serial port %d (irq %d)...\n", DRVNAME, info->line,
		   state->irq);
	
	DGLOCK(flags);

	/*
	 * clear delta_msr_wait queue to avoid mem leaks: we may free the irq
	 * here so the queue might never be waken up
	 */
	wake_up_interruptible(&info->delta_msr_wait);
	
	/*
	 * First unlink the serial port from the IRQ chain...
	 */
	if (info->next_port)
		info->next_port->prev_port = info->prev_port;
	if (info->prev_port)
		info->prev_port->next_port = info->next_port;
	else
		IRQ_ports[state->irq] = info->next_port;
	figure_IRQ_timeout(state->irq);
	
	/*
	 * Free the IRQ, if necessary
	 */
	if (state->irq && (!IRQ_ports[state->irq] ||
			  !IRQ_ports[state->irq]->next_port)) {
		if (IRQ_ports[state->irq]) {
			free_irq(state->irq, &IRQ_ports[state->irq]);
			retval = request_irq(state->irq, cpci_interrupt,
						 SA_SHIRQ, "ClassicPCI", &IRQ_ports[state->irq]);
			
			if (retval)
				printk("%s: serial shutdown: request_irq: error %d"
					   "  Couldn't reacquire IRQ.\n", DRVNAME, retval);
		} else
			free_irq(state->irq, &IRQ_ports[state->irq]);
	}

	if (info->xmit_buf) {
		free_page((unsigned long) info->xmit_buf);
		info->xmit_buf = 0;
	}
	info->IER = 0;
	serial_out(info, UART_IER, 0x00);	/* disable all intrs */
	info->MCR &= ~UART_MCR_OUT2;
	
	/* disable break condition */
	serial_out(info, UART_LCR, serial_in(info, UART_LCR) & ~UART_LCR_SBC);
	
	if (!info->tty || (info->tty->termios->c_cflag & HUPCL))
		info->MCR &= ~(UART_MCR_DTR|UART_MCR_RTS);

	serial_out(info, UART_MCR, info->MCR);

	/* disable FIFO's */	
	serial_out(info, UART_FCR, (UART_FCR_CLEAR_RCVR |
					 UART_FCR_CLEAR_XMIT));
	(void)serial_in(info, UART_RX);    /* read data port to reset things */
	
	if (info->tty)
		set_bit(TTY_IO_ERROR, &info->tty->flags);

	if (uart_config[info->state->type].flags & UART_STARTECH) {
		
		if(info->state->type == PORT_17158) {
			/* Dont do *ANYTHING* here. */
		} else {
		/* Arrange to enter sleep mode */
			serial_out(info, UART_LCR, 0xBF);
			serial_out(info, UART_EFR, UART_EFR_ECB);
			serial_out(info, UART_IER, UART_IERX_SLEEP);
		}
	}
	if (info->state->type == PORT_16750) {
		/* Arrange to enter sleep mode */
		serial_out(info, UART_IER, UART_IERX_SLEEP);
	}
	info->flags &= ~ASYNC_INITIALIZED;
	DGUNLOCK(flags);
}

static int size_fifo(struct cpci_async_struct *info);

/*
 * This routine is called to set the UART divisor registers to match
 * the specified baud rate for a serial port.
 */
static void change_speed(struct cpci_async_struct *info,
			 struct termios *old_termios)
{
	unsigned short port;
	int	quot = 0, baud_base, baud;
	unsigned cflag, cval, fcr = 0;
	int	bits;
	unsigned long	flags;
	int IER = 0;
	unsigned char tmp = 0;	

	if (!info->tty || !info->tty->termios)
		return;
	cflag = info->tty->termios->c_cflag;


	if(!CONFIGURED_SERIAL_PORT(info))
		return;
		
	if(info->port) port = info->port;
	
	IER = serial_in(info, UART_IER);

	serial_out(info, UART_IER, 0);
	
	/* byte size and parity */
	switch (cflag & CSIZE) {
		  case CS5: cval = 0x00; bits = 7; break;
		  case CS6: cval = 0x01; bits = 8; break;
		  case CS7: cval = 0x02; bits = 9; break;
		  case CS8: cval = 0x03; bits = 10; break;
		  /* Never happens, but GCC is too dumb to figure it out */
		  default:  cval = 0x00; bits = 7; break;
		  }
	if (cflag & CSTOPB) {
		cval |= 0x04;
		bits++;
	}
	if (cflag & PARENB) {
		cval |= UART_LCR_PARITY;
		bits++;
	}
	if (!(cflag & PARODD))
		cval |= UART_LCR_EPAR;
#ifdef CMSPAR
	if (cflag & CMSPAR)
		cval |= UART_LCR_SPAR;
#endif

	/* Determine divisor based on baud rate */
	baud = tty_get_baud_rate(info->tty);
	if (!baud)
		baud = 9600;	/* B0 transition handled in cpci_set_termios */
	
	baud_base = info->state->baud_base;
	if (info->state->type == PORT_16C950) {
		if (baud <= baud_base)
			serial_icr_write(info, UART_TCR, 0);
		else if (baud <= 2*baud_base) {
			serial_icr_write(info, UART_TCR, 0x8);
			baud_base = baud_base * 2;
		} else if (baud <= 4*baud_base) {
			serial_icr_write(info, UART_TCR, 0x4);
			baud_base = baud_base * 4;
		} else
			serial_icr_write(info, UART_TCR, 0);
	}
	
	if (baud == 38400 &&
		((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_CUST))
		quot = info->state->custom_divisor;
	else {
		if (baud == 134)
			/* Special case since 134 is really 134.5 */
			quot = (2*baud_base / 269);
		else if (baud)
			quot = baud_base / baud;
	}
	/* If the quotient is zero refuse the change */
	if (!quot && old_termios) {
		info->tty->termios->c_cflag &= ~CBAUD;
		info->tty->termios->c_cflag |= (old_termios->c_cflag & CBAUD);
		baud = tty_get_baud_rate(info->tty);
		if (!baud)
			baud = 9600;
		if (baud == 38400 &&
			((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_CUST))
			quot = info->state->custom_divisor;
		else {
			if (baud == 134)
				/* Special case since 134 is really 134.5 */
				quot = (2*baud_base / 269);
			else if (baud)
				quot = baud_base / baud;
		}
	}
	/* As a last resort, if the quotient is zero, default to 9600 bps */
	if (!quot)
		quot = baud_base / 9600;
	
	/*
	 * Work around a bug in the Oxford Semiconductor 952 rev B
	 * chip which causes it to seriously miscalculate baud rates
	 * when DLL is 0.
	 */
	if (((quot & 0xFF) == 0) && (info->state->type == PORT_16C950) &&
	(info->state->revision == 0x5201))
		quot++;
	
	info->quot = quot;
	info->timeout = ((info->xmit_fifo_size*HZ*bits*quot) / baud_base);
	info->timeout += HZ/50;		/* Add .02 seconds of slop */

	/* Set up FIFO's */
	if (uart_config[info->state->type].flags & UART_USE_FIFO) {

		serial_out(info, UART_FCR, UART_FCR_ENABLE_FIFO);

		if(info->state->type == PORT_16654)
			fcr = UART_FCR_ENABLE_FIFO | UART_FCR_CLEAR_RCVR | UART_16654_FCR_RXTRIGGER_16;
		else
			fcr = UART_FCR_ENABLE_FIFO | UART_FCR_CLEAR_RCVR | UART_FCR_TRIGGER_8;
					
	}
	if (info->state->type == PORT_16750)
		fcr |= UART_FCR7_64BYTE;
	
	/* CTS flow control flag and modem status interrupts */
	info->IER &= ~UART_IER_MSI;

	if (cflag & CRTSCTS || info->digiext.digi_flags & RTSPACE || info->digiext.digi_flags & CTSPACE) {
		info->flags |= ASYNC_CTS_FLOW;
		info->IER |= UART_IER_MSI;
	} else
{
		info->flags &= ~ASYNC_CTS_FLOW;
}
	if (cflag & CLOCAL)
		info->flags &= ~ASYNC_CHECK_CD;
	else {
		info->flags |= ASYNC_CHECK_CD;
		info->IER |= UART_IER_MSI;
	}
	serial_out(info, UART_IER, info->IER);

	/*
	 * Set up parity check flag
	 */
#define RELEVANT_IFLAG(iflag) (iflag & (IGNBRK|BRKINT|IGNPAR|PARMRK|INPCK))
	
	info->read_status_mask = UART_LSR_OE | UART_LSR_THRE | UART_LSR_DR;
	if (I_INPCK(info->tty))
		info->read_status_mask |= UART_LSR_FE | UART_LSR_PE;
	if (I_BRKINT(info->tty) || I_PARMRK(info->tty))
		info->read_status_mask |= UART_LSR_BI;
	
	/*
	 * Characters to ignore
	 */
	info->ignore_status_mask = 0;
	if (I_IGNPAR(info->tty))
		info->ignore_status_mask |= UART_LSR_PE | UART_LSR_FE;
	if (I_IGNBRK(info->tty)) {
		info->ignore_status_mask |= UART_LSR_BI;
		/*
		 * If we're ignore parity and break indicators, ignore
		 * overruns too.  (For real raw support).
		 */
		if (I_IGNPAR(info->tty))
			info->ignore_status_mask |= UART_LSR_OE;
	}
	/*
	 * !!! ignore all characters if CREAD is not set
	 */
	if ((cflag & CREAD) == 0)
		info->ignore_status_mask |= UART_LSR_DR;
	DGLOCK(flags);
	if (uart_config[info->state->type].flags & UART_STARTECH) {
		
		/* This needs to be set in *ALL* cases, if the chip is a. */
		/* 17C158.						  */
		
		if(info->state->type == PORT_17158)
			tmp |= UART_EFR_ECB;
			
		/* in all STARTECH UARTs (such as the Exar 650, 654, 854 */
		/* and 158) - loopback mode using CTS hardware flow 	 */
		/* control *WILL FAIL*, considering that CTS is ALWAYS 	 */
		/* reported high, however, internal to the UART - it is	 */
		/* gated low and causes NO data to EVER be transmitted.  */
		
		if(cflag & CRTSCTS  || info->digiext.digi_flags & RTSPACE || info->digiext.digi_flags & CTSPACE) {
			if(info->loopedback)
				tmp |= UART_EFR_RTS;
			else
				tmp |= UART_EFR_RTS | UART_EFR_CTS;
		}
				
		if(info->state->type == PORT_17158)
			serial_out(info, UART_EXTENDED_EFR, tmp);
		else {
			serial_out(info, UART_LCR, 0xBF);
			serial_out(info, UART_EFR, tmp);
		}
	
	}
	
	serial_out(info, UART_LCR, cval | UART_LCR_DLAB);	/* set DLAB */
	serial_out(info, UART_DLL, quot & 0xff);	/* LS of divisor */
	serial_out(info, UART_DLM, quot >> 8);		/* MS of divisor */
	if (info->state->type == PORT_16750)
		serial_out(info, UART_FCR, fcr); 	/* set fcr */
	serial_out(info, UART_LCR, cval);		/* reset DLAB */
	info->LCR = cval;                               /* Save LCR */
	
	if (info->state->type != PORT_16750) {
		if (fcr & UART_FCR_ENABLE_FIFO) {
			serial_out(info, UART_FCR, UART_FCR_ENABLE_FIFO);
		}
			
		serial_out(info, UART_FCR, fcr); 	/* set fcr */
	}
	
	DGUNLOCK(flags);
}

static void cpci_put_char(struct tty_struct *tty, unsigned char ch)
{
	struct cpci_async_struct *info = (struct cpci_async_struct *)tty->driver_data;
	unsigned long flags;

	if (serial_paranoia_check(info, tty->device, "cpci_put_char"))
		return;

	if (!tty || !info->xmit_buf)
		return;

	DGLOCK(flags);
	if (info->xmit_cnt >= CPCI_XMIT_SIZE - 1) {
		DGUNLOCK(flags);
		return;
	}

	info->xmit_buf[info->xmit_head++] = ch;
	info->xmit_head &= CPCI_XMIT_SIZE-1;
	info->xmit_cnt++;
	DGUNLOCK(flags);
}

static void cpci_flush_chars(struct tty_struct *tty)
{
	struct cpci_async_struct *info = (struct cpci_async_struct *)tty->driver_data;
	unsigned long flags;
	
	if(tty->driver.subtype == SERIAL_TYPE_INFO)
		return;
	
	if (serial_paranoia_check(info, tty->device, "cpci_flush_chars"))
		return;

	if (info->xmit_cnt <= 0 || tty->stopped || tty->hw_stopped ||
		!info->xmit_buf){
		return;
	}

	DGLOCK(flags);
	info->IER |= UART_IER_THRI;
	serial_out(info, UART_IER, info->IER);
	DGUNLOCK(flags);
}

static int cpci_write(struct tty_struct * tty, int from_user,
			const unsigned char *buf, int count)
{
	int	c, ret = 0;
	struct cpci_async_struct *info = (struct cpci_async_struct *)tty->driver_data;
	unsigned long flags;

	if(tty->driver.subtype == SERIAL_TYPE_INFO)
		return (0);
				
	if (serial_paranoia_check(info, tty->device, "cpci_write"))
		return 0;

	if (!tty || !info->xmit_buf || !tmp_buf)
		return 0;
	
	save_flags(flags);
	//DGLOCK(flags);
	if (from_user) {
		down(&tmp_buf_sem);
		while (1) {
			c = MIN(count,
				MIN(CPCI_XMIT_SIZE - info->xmit_cnt - 1,
					CPCI_XMIT_SIZE - info->xmit_head));
			if (c <= 0)
				break;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
			c -= copy_from_user(tmp_buf, buf, c);
			
			if (!c) {
				if (!ret)
					ret = -EFAULT;
				
				break;
			}
#else
			copy_from_user(tmp_buf, buf, c);
#endif		
			cli();
			//DGLOCK(flags);
			c = MIN(c, MIN(CPCI_XMIT_SIZE - info->xmit_cnt - 1,
					   CPCI_XMIT_SIZE - info->xmit_head));
			memcpy(info->xmit_buf + info->xmit_head, tmp_buf, c);
			info->xmit_head = ((info->xmit_head + c) &
					   (CPCI_XMIT_SIZE-1));

				
			info->xmit_cnt += c;
			DGUNLOCK(flags);
			buf += c;
			count -= c;
			ret += c;
		}
		up(&tmp_buf_sem);
	} else {
		while (1) {
			//DGLOCK(flags);
			cli();
			c = MIN(count,
				MIN(CPCI_XMIT_SIZE - info->xmit_cnt - 1,
					CPCI_XMIT_SIZE - info->xmit_head));
			if (c <= 0) {
				DGUNLOCK(flags);
				break;
			}
			memcpy(info->xmit_buf + info->xmit_head, buf, c);
			info->xmit_head = ((info->xmit_head + c) &
					   (CPCI_XMIT_SIZE-1));
			info->xmit_cnt += c;
			DGUNLOCK(flags);
			buf += c;
			count -= c;
			ret += c;
		}
	}

	if (info->xmit_cnt && !tty->stopped && !tty->hw_stopped &&
		!(info->IER & UART_IER_THRI))
	{
		info->IER |= UART_IER_THRI;
		serial_out(info, UART_IER, info->IER);
	}
			
	return ret;
}

static int cpci_write_room(struct tty_struct *tty)
{
	struct cpci_async_struct *info = (struct cpci_async_struct *)tty->driver_data;
	int	ret;
				
	if(tty->driver.subtype == SERIAL_TYPE_INFO)
		return(0);
		
	if (serial_paranoia_check(info, tty->device, "cpci_write_room"))
		return 0;
	ret = CPCI_XMIT_SIZE - info->xmit_cnt - 1;
	if (ret < 0)
		ret = 0;
	return ret;
}

static int cpci_chars_in_buffer(struct tty_struct *tty)
{
	struct cpci_async_struct *info = (struct cpci_async_struct *)tty->driver_data;
	
	if(tty->driver.subtype == SERIAL_TYPE_INFO)
		return(0);
		
	if (serial_paranoia_check(info, tty->device, "cpci_chars_in_buffer"))
		return 0;
	return info->xmit_cnt;
}

static void cpci_flush_buffer(struct tty_struct *tty)
{
	struct cpci_async_struct *info = (struct cpci_async_struct *)tty->driver_data;
	unsigned long flags;
	
	if(tty->driver.subtype == SERIAL_TYPE_INFO)
		return;

	if (serial_paranoia_check(info, tty->device, "cpci_flush_buffer"))
		return;
	
	DGLOCK(flags);
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
	DGUNLOCK(flags);
	wake_up_interruptible(&tty->write_wait);
	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		tty->ldisc.write_wakeup)
		(tty->ldisc.write_wakeup)(tty);
}

/*
 * This function is used to send a high-priority XON/XOFF character to
 * the device
 */
static void cpci_send_xchar(struct tty_struct *tty, char ch)
{
	struct cpci_async_struct *info = (struct cpci_async_struct *)tty->driver_data;

	if(tty->driver.subtype == SERIAL_TYPE_INFO)
		return;
	
	if (serial_paranoia_check(info, tty->device, "cpci_send_char"))
		return;

	info->x_char = ch;
	if (ch) {
		/* Make sure transmit interrupts are on */
		info->IER |= UART_IER_THRI;
		serial_out(info, UART_IER, info->IER);
	}
}

/*
 * ------------------------------------------------------------
 * cpci_throttle()
 *
 * This routine is called by the upper-layer tty layer to signal that
 * incoming characters should be throttled.
 * ------------------------------------------------------------
 */
static void cpci_throttle(struct tty_struct * tty)
{
	struct cpci_async_struct *info = (struct cpci_async_struct *)tty->driver_data;
	unsigned long flags;
	
	if(tty->driver.subtype == SERIAL_TYPE_INFO)
		return;
		
	DEBUG_PRINTK("%s: (minor %d) throttle : %d....\n", DRVNAME, MINOR(tty->device),
		tty->ldisc.chars_in_buffer(tty));

	if (serial_paranoia_check(info, tty->device, "cpci_throttle"))
	 	return;
	
	if (I_IXOFF(tty))
		cpci_send_xchar(tty, STOP_CHAR(tty));

	if (tty->termios->c_cflag & CRTSCTS  || info->digiext.digi_flags & RTSPACE || info->digiext.digi_flags & CTSPACE) {
		info->MCR &= ~UART_MCR_RTS;
	}			

	DGLOCK(flags);
	serial_out(info, UART_MCR, info->MCR);
	info->throttled = 1;
	DGUNLOCK(flags);
}

static void cpci_unthrottle(struct tty_struct * tty)
{
	struct cpci_async_struct *info = (struct cpci_async_struct *)tty->driver_data;
	unsigned long flags;
	
	if(tty->driver.subtype == SERIAL_TYPE_INFO)
		return;

	DEBUG_PRINTK("%s: (minor %d) unthrottle : %d....\n", DRVNAME, MINOR(tty->device),
		tty->ldisc.chars_in_buffer(tty));

	/* If the tty is closing, don't bother turning RTS back on. The */
	/* only thing we'll succeed in doing then, is turning RTS on 	*/
	/* infinitely :)						*/
	
	if(tty->closing) {
		info->throttled = 0;
		return;
	}
	
	if (serial_paranoia_check(info, tty->device, "cpci_unthrottle")) {
		info->throttled = 0;
		return;
	}
	
	if (I_IXOFF(tty)) {
		if (info->x_char)
			info->x_char = 0;
		else
			cpci_send_xchar(tty, START_CHAR(tty));
	}
	if (tty->termios->c_cflag & CRTSCTS  || info->digiext.digi_flags & RTSPACE || info->digiext.digi_flags & CTSPACE) {
		info->MCR |= UART_MCR_RTS;
	}
	
	info->throttled = 0;
	DGLOCK(flags);
	serial_out(info, UART_MCR, info->MCR);
	
	DGUNLOCK(flags);
}

/*
 * ------------------------------------------------------------
 * cpci_ioctl() and friends
 * ------------------------------------------------------------
 */


static int get_serial_info(struct cpci_async_struct * info,
			   struct serial_struct * retinfo)
{
	struct serial_struct tmp;
	struct cpci_serial_state *state = info->state;

	if (!retinfo)
		return -EFAULT;
	memset(&tmp, 0, sizeof(tmp));
	tmp.type = state->type;
	tmp.line = state->line;
	tmp.port = state->port;
	tmp.iomem_base = state->iomem_base;
	tmp.io_type = state->io_type;
	tmp.io_flags = state->io_flags;
	tmp.irq = state->irq;
	tmp.flags = state->flags;
	tmp.xmit_fifo_size = state->xmit_fifo_size;
	tmp.baud_base = state->baud_base;
	tmp.close_delay = state->close_delay;
	tmp.closing_wait = state->closing_wait;
	tmp.custom_divisor = state->custom_divisor;
	tmp.hub6 = state->hub6;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
	copy_to_user(retinfo, &tmp, sizeof(*retinfo));
#else	
	if (copy_to_user(retinfo,&tmp,sizeof(*retinfo)))
		return -EFAULT;
#endif
	return 0;
}

static int set_serial_info(struct cpci_async_struct * info,
			   struct serial_struct * new_info)
{
	struct serial_struct new_serial;
 	struct cpci_serial_state old_state, *state;
	unsigned int		i,change_irq,change_port;
	int 			retval = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
	copy_from_user(&new_serial, new_info, sizeof(new_serial));
#else
	if (copy_from_user(&new_serial,new_info,sizeof(new_serial)))
		return -EFAULT;
#endif

	state = info->state;
	old_state = *state;

	change_irq = new_serial.irq != state->irq;
	change_port = (new_serial.port != state->port) ||
		(new_serial.hub6 != state->hub6);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
	if (!suser())
#else
	if (!capable(CAP_SYS_ADMIN))
#endif
	{
		if (change_irq || change_port ||
			(new_serial.baud_base != state->baud_base) ||
			(new_serial.type != state->type) ||
			(new_serial.close_delay != state->close_delay) ||
			(new_serial.xmit_fifo_size != state->xmit_fifo_size) ||
			((new_serial.flags & ~ASYNC_USR_MASK) !=
			 (state->flags & ~ASYNC_USR_MASK)))
			return -EPERM;
		state->flags = ((state->flags & ~ASYNC_USR_MASK) |
				   (new_serial.flags & ASYNC_USR_MASK));
		info->flags = ((info->flags & ~ASYNC_USR_MASK) |
				   (new_serial.flags & ASYNC_USR_MASK));
		state->custom_divisor = new_serial.custom_divisor;
		goto check_and_exit;
	}

	new_serial.irq = irq_cannonicalize(new_serial.irq);

	if ((new_serial.irq >= NR_IRQS) || (new_serial.port > 0xffff) ||
		(new_serial.baud_base < 9600)|| (new_serial.type < PORT_UNKNOWN) ||
		(new_serial.type > PORT_MAX) || (new_serial.type == PORT_CIRRUS) ||
		(new_serial.type == PORT_STARTECH)) {
		return -EINVAL;
	}

	if ((new_serial.type != state->type) ||
		(new_serial.xmit_fifo_size <= 0))
		new_serial.xmit_fifo_size =
			uart_config[new_serial.type].dfl_xmit_fifo_size;

	/* Make sure address is not already in use */
	if (new_serial.type) {
		for (i = 0 ; i < NR_PORTS; i++) {

			if(cpci_table[i].io_type == CPCI_IO_MEM) {
				if( (state != &cpci_table[i]) &&
				(cpci_table[i].iomem_base == new_serial.iomem_base) )
					return -EADDRINUSE;
			} else {
				if( (state != &cpci_table[i]) &&
				(cpci_table[i].port == new_serial.port) &&
					(cpci_table[i].type) )
						return -EADDRINUSE;
			}
		}
	}

	if((change_port || change_irq) && (state->count > 1))
		return -EBUSY;

	/*
	 * OK, past this point, all the error checking has been done.
	 * At this point, we start making changes.....
	 */

	state->baud_base = new_serial.baud_base;
	state->flags = ((state->flags & ~ASYNC_FLAGS) |
			(new_serial.flags & ASYNC_FLAGS));
	info->flags = ((state->flags & ~ASYNC_INTERNAL_FLAGS) |
			   (info->flags & ASYNC_INTERNAL_FLAGS));
	state->custom_divisor = new_serial.custom_divisor;
	state->type = new_serial.type;
	state->close_delay = new_serial.close_delay * HZ/100;
	state->closing_wait = new_serial.closing_wait * HZ/100;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	info->tty->low_latency = (info->flags & ASYNC_LOW_LATENCY) ? 1 : 0;
#endif
	info->xmit_fifo_size = state->xmit_fifo_size =
		new_serial.xmit_fifo_size;

	if(state->port)
		release_region(state->port,8);
	if (change_port || change_irq) {
		/*
		 * We need to shutdown the serial port at the old
		 * port/irq combination.
		 */
		shutdown(info);
		state->irq = new_serial.irq;
		info->port = state->port = new_serial.port;
		info->hub6 = state->hub6 = new_serial.hub6;
	}
	if (state->type != PORT_UNKNOWN)
		if(state->port)
			request_region(state->port,8,"serial(set)");

	
check_and_exit:

	if(!CONFIGURED_SERIAL_PORT(state))
		return(-1);

	if (info->flags & ASYNC_INITIALIZED) {
		if (((old_state.flags & ASYNC_SPD_MASK) !=
			 (state->flags & ASYNC_SPD_MASK)) ||
			(old_state.custom_divisor != state->custom_divisor)) {
			change_speed(info, 0);
		}
	} else
		retval = startup(info);
	return retval;
}

/*
 * get_lsr_info - get line status register info
 *
 * Purpose: Let user call ioctl() to get info when the UART physically
 * 	    is emptied.  On bus types like RS485, the transmitter must
 * 	    release the bus after transmitting. This must be done when
 * 	    the transmit shift register is empty, not be done when the
 * 	    transmit holding register is empty.  This functionality
 * 	    allows an RS485 driver to be written in user space.
 */
static int get_lsr_info(struct cpci_async_struct * info, unsigned int *value)
{
	unsigned char status;
	unsigned int result;
	unsigned long flags;

	DGLOCK(flags);
	status = serial_in(info, UART_LSR);
	DGUNLOCK(flags);
	result = ((status & UART_LSR_TEMT) ? TIOCSER_TEMT : 0);
	/*return putUser(result,value);*/
	putUser(result,value);
	return(0);
}


/* If "value" is NULL, return result to caller.  Otherwise use value as
 * a userspace address for the result.
 */
static int get_modem_info(struct cpci_async_struct * info, unsigned int *value )
{
	unsigned char control, status;
	unsigned int result;
	unsigned long flags;
	
	control = info->MCR;
	DGLOCK(flags);
	status = serial_in(info, UART_MSR);
	DGUNLOCK(flags);
	result =  ((control & UART_MCR_RTS) ? TIOCM_RTS : 0)
		| ((control & UART_MCR_DTR) ? TIOCM_DTR : 0)
#ifdef TIOCM_OUT1
		| ((control & UART_MCR_OUT1) ? TIOCM_OUT1 : 0)
		| ((control & UART_MCR_OUT2) ? TIOCM_OUT2 : 0)
#endif
		/*| ((status  & UART_MSR_DCD) ? TIOCM_CAR : 0)*/
		| (info->digiext.digi_flags & DIGI_ALTPIN?
			((status  & UART_MSR_DSR) ? TIOCM_CAR : 0) :
			((status  & UART_MSR_DCD) ? TIOCM_CAR : 0))
		| ((status  & UART_MSR_RI) ? TIOCM_RNG : 0)
		/*| ((status  & UART_MSR_DSR) ? TIOCM_DSR : 0)*/
		| (info->digiext.digi_flags & DIGI_ALTPIN?
			((status  & UART_MSR_DCD) ? TIOCM_DSR : 0) :
			((status  & UART_MSR_DSR) ? TIOCM_DSR : 0))
		| ((status  & UART_MSR_CTS) ? TIOCM_CTS : 0);
	if( value ) {
		putUser(result,value);
		return(0);
	} else
		return( result );
}

static int set_modem_info(struct cpci_async_struct * info, unsigned int cmd,
			  unsigned int *value)
{
	int error;
	unsigned int arg;
	unsigned long flags;

	error = getUser(arg, value);
	if (error)
		return error;
	switch (cmd) {
	case TIOCMBIS:
		if (arg & TIOCM_RTS)
			info->MCR |= UART_MCR_RTS;
		if (arg & TIOCM_DTR)
			info->MCR |= UART_MCR_DTR;
#ifdef TIOCM_OUT1
		if (arg & TIOCM_OUT1)
			info->MCR |= UART_MCR_OUT1;
		if (arg & TIOCM_OUT2)
			info->MCR |= UART_MCR_OUT2;
#endif
		break;
	case TIOCMBIC:
		if (arg & TIOCM_RTS)
			info->MCR &= ~UART_MCR_RTS;
		if (arg & TIOCM_DTR)
			info->MCR &= ~UART_MCR_DTR;
#ifdef TIOCM_OUT1
		if (arg & TIOCM_OUT1)
			info->MCR &= ~UART_MCR_OUT1;
		if (arg & TIOCM_OUT2)
			info->MCR &= ~UART_MCR_OUT2;
#endif
		break;
	case TIOCMSET:
		/* First, set the local copy of the MCR to some defaults.. 	*/
		/* (No RTS/DTR, and no OUT1/OUT2, only if applicable)		*/

#ifdef TIOCM_OUT1
		info->MCR &= ~UART_MCR_OUT1;
		info->MCR &= ~UART_MCR_OUT2;
#endif
		info->MCR &= ~UART_MCR_RTS;
		info->MCR &= ~UART_MCR_DTR;

		
		/* Now, set these values to that which the user requested. */
		
		if(arg & TIOCM_RTS)
			info->MCR |= UART_MCR_RTS;
		if(arg & TIOCM_DTR)
			info->MCR |= UART_MCR_DTR;
#ifdef TIOCM_OUT1
		if(arg & TIOCM_OUT1)
			info->MCR |= UART_MCR_OUT1;
		if(arg & TIOCM_OUT2)
			info->MCR |= UART_MCR_OUT2;
#endif
			
		 /* Hack to allow loopback to stay, when this ioctl */
		 /* is called - CWS					*/

		if(info->loopedback)
			info->MCR |= UART_MCR_LOOP;

		break;
	default:
		return -EINVAL;
	}
	DGLOCK(flags);
	serial_out(info, UART_MCR, info->MCR);
	DGUNLOCK(flags);
	return 0;
}

static int do_autoconfig(struct cpci_async_struct * info)
{
	int			retval;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
	if(!suser())
#else
	if (!capable(CAP_SYS_ADMIN))
#endif
		return -EPERM;
	
	if (info->state->count > 1)
		return -EBUSY;
	
	shutdown(info);
	autoconfig(info->state);

	retval = startup(info);
	if (retval)
		return retval;
	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
/*
 * cpci_break() --- routine which turns the break handling on or off
 */
static void cpci_break(struct tty_struct *tty, int break_state)
{
	struct cpci_async_struct * info = (struct cpci_async_struct *)tty->driver_data;
	unsigned long flags;
	
	if (serial_paranoia_check(info, tty->device, "cpci_break"))
		return;

	if(!CONFIGURED_SERIAL_PORT(info))
		return;
		
	DGLOCK(flags);
	if (break_state == -1)
		serial_out(info, UART_LCR,
			   serial_in(info, UART_LCR) | UART_LCR_SBC);
	else
		serial_out(info, UART_LCR,
			   serial_in(info, UART_LCR) & ~UART_LCR_SBC);
	DGUNLOCK(flags);
}

#endif

static int cpci_ioctl(struct tty_struct *tty, struct file * file,
			unsigned int cmd, unsigned long arg)
{
	int error;
	struct cpci_async_struct * info = (struct cpci_async_struct *)tty->driver_data;
	struct cpci_async_icount cprev, cnow;	/* kernel counter temps */
	struct serial_icounter_struct *p_cuser;	/* user space */
	unsigned long flags;

	error = 0;
	
	/* Management minor stuff */
	
	if(tty->driver.subtype == SERIAL_TYPE_INFO) {
		switch(cmd) {

			case DIGI_TIMERIRQ:
			{
				getUser(cpci_use_rs_timer, arg);
				DEBUG_PRINTK( "%s: %s RS_TIMER use\n",
					DRVNAME, cpci_use_rs_timer? "Enabling" : "Disabling" );
				break;
			}

			case DIGI_GETDD:
			{
				/* This ioctl returns the number of boards */
				/* currently configured in the system	   */
				struct digi_dinfo ddi;
				ddi.dinfo_nboards = cpci_totalboards;
				sprintf(ddi.dinfo_version, "%s %s", cpci_name, cpci_version);
				copy_to_user((char*)arg, &ddi, sizeof(ddi));
				break;
			}
			
			case DIGI_GETBD:
			{
				int brd;
				struct digi_info di;
				
				getUser(brd, arg);
				
				if((brd < 0) || (brd >= cpci_totalboards) || (cpci_totalboards == 0))
					return(-ENODEV);
					
				memset(&di, 0, sizeof(di));
				
				di.info_bdnum = brd;
				di.info_bdstate = board[brd].dpastatus;
				di.info_bdtype = board[brd].dpatype;
				di.info_nports = board[brd].numports;
				di.info_ioport = board[brd].iobase;
				di.info_physaddr = board[brd].membase;
				di.info_physsize = 4096;
				
							/* First port on board is at ports[di.firstchan]*/
				di.firstchan = board[brd].firstchan;
				
				copy_to_user((char *)arg, &di, sizeof(di));
				break;
			}
					
			default:
				printk("%s: cpci_ioctl(): Unimplemented management ioctl call\n", DRVNAME );
				return(-EINVAL);
		}
		return(0);
	}
	
	if (serial_paranoia_check(info, tty->device, "cpci_ioctl"))
		return -ENODEV;

	if ((cmd != TIOCGSERIAL) && (cmd != TIOCSSERIAL) &&
		(cmd != TIOCSERCONFIG) && (cmd != TIOCSERGSTRUCT) &&
		(cmd != TIOCMIWAIT) && (cmd != TIOCGICOUNT)) {
		if (tty->flags & (1 << TTY_IO_ERROR))
				return -EIO;
	}
	
	switch (cmd) {
		case DIGI_GETA:
		{
				if(info->tty->termios->c_cflag & CRTSCTS)
				{
				    info->digiext.digi_flags |= (RTSPACE | CTSPACE);
				}
				/* FIFO triggers only available on Neoclassics */
				if(info->state->type != PORT_17158)
					info->digiext.rxtrg = info->digiext.txtrg = 0;

				copy_to_user((char*)arg, &info->digiext, sizeof(digi_t));
				break;
		}

		case DIGI_SETAW:
		case DIGI_SETAF:
		if ((cmd) == (DIGI_SETAW)) {
			tty_wait_until_sent(tty, 0);
		} else {
			if (tty->ldisc.flush_buffer)
				tty->ldisc.flush_buffer(tty);
		}
		/* Fall Thru */

		case DIGI_SETA:
		{
			copy_from_user(&info->digiext, (char*)arg, sizeof(digi_t));
			
			if((info->digiext.digi_flags & RTSPACE) || (info->digiext.digi_flags & CTSPACE))
			{
				info->flags |= ASYNC_CTS_FLOW;
				info->tty->termios->c_cflag |= CRTSCTS;
			}
			else
			{
				info->flags &= ~ASYNC_CTS_FLOW;
				info->tty->termios->c_cflag &= ~CRTSCTS;
			}

			if( info->digiext.rxtrg )
			{
				if( info->state->type == PORT_17158)
					serial_out(info, UART_EXTENDED_RXTRG, info->digiext.rxtrg );
				else
					printk("%s: Settable RXTRG levels supported only on 17X58 UARTs\n", DRVNAME );
			}
			if( info->digiext.txtrg )
			{
				if( info->state->type == PORT_17158)
					serial_out(info, UART_EXTENDED_TXTRG, info->digiext.txtrg );
				else
					printk("%s: Settable TXTRG levels supported only on 17X58 UARTs, port:%d\n", DRVNAME, MINOR(info->tty->device) );

			}

			break;
		}

		case DIGI_GET_NI_INFO:
		{
			struct ni_info niinfo;
			memset(&niinfo, 0, sizeof(struct ni_info));
				
			niinfo.mstat = get_modem_info(info, NULL );
			niinfo.dtr = niinfo.mstat & TIOCM_DTR;
			niinfo.rts = niinfo.mstat & TIOCM_RTS;
			niinfo.cts = niinfo.mstat & TIOCM_CTS;
			niinfo.ri = niinfo.mstat & TIOCM_RI;

			/* get_modem_info takes care of ALTPIN */
			niinfo.dcd = niinfo.mstat & TIOCM_CAR;

			niinfo.iflag = info->tty->termios->c_iflag;
			niinfo.oflag = info->tty->termios->c_oflag;
			niinfo.cflag = info->tty->termios->c_cflag;
			niinfo.lflag = info->tty->termios->c_lflag;
			niinfo.hflow = info->flags & ASYNC_CTS_FLOW? 1 : 0;
			niinfo.recv_stopped = info->throttled ? 1 : 0;
			niinfo.xmit_stopped = (info->tty->hw_stopped
				|| info->tty->stopped)? 1 : 0;

			niinfo.curtx = info->xmit_cnt;
			niinfo.currx = info->tty->read_cnt;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
			copy_to_user((struct ni_info *)arg, &niinfo, sizeof(struct ni_info));
#else
			if (copy_to_user((struct ni_info *)arg, &niinfo,
				sizeof(struct ni_info))) {
				return -EFAULT;
			}
#endif
			break;

		}

		case DIGI_LOOPBACK:
		{
			int loopback;
			getUser(loopback, arg);
			/* Enable/disable internal loopback for this port */
			info->loopedback = loopback;
			if( loopback )
	 			info->MCR |= UART_MCR_LOOP;
			else
	 			info->MCR &= ~UART_MCR_LOOP;
			serial_out(info, UART_MCR, info->MCR);
			break;
		}


		case TIOCMGET:
			return get_modem_info(info, (unsigned int *) arg);
		case TIOCMBIS:
		case TIOCMBIC:
		case TIOCMSET:
			return set_modem_info(info, cmd, (unsigned int *) arg);
		case TIOCGSERIAL:
			return get_serial_info(info,
						   (struct serial_struct *) arg);
		case TIOCSSERIAL:
			return set_serial_info(info,
						   (struct serial_struct *) arg);
		case TIOCSERCONFIG:
			return do_autoconfig(info);

		case TIOCSERGETLSR: /* Get line status register */
			return get_lsr_info(info, (unsigned int *) arg);

		case TIOCSERGSTRUCT:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
			copy_to_user((struct cpci_async_struct *)arg,
					info, sizeof(struct cpci_async_struct));
#else
			if (copy_to_user((struct cpci_async_struct *) arg,
					 info, sizeof(struct cpci_async_struct)))
				return -EFAULT;
#endif

			return 0;
				
		/*
		 * Wait for any of the 4 modem inputs (DCD,RI,DSR,CTS) to change
		 * - mask passed in arg for lines of interest
 		 *   (use |'ed TIOCM_RNG/DSR/CD/CTS for masking)
		 * Caller should use TIOCGICOUNT to see which one it was
		 */
		case TIOCMIWAIT:
			DGLOCK(flags);
			/* note the counters on entry */
			cprev = info->state->icount;
			DGUNLOCK(flags);
			while (1) {
				interruptible_sleep_on(&info->delta_msr_wait);
				/* see if a signal did it */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
				if (current->signal & ~current->blocked)
				                        return -ERESTARTSYS;
#else
				if (signal_pending(current))
					return -ERESTARTSYS;
#endif

				DGLOCK(flags);
				cnow = info->state->icount; /* atomic copy */
				DGUNLOCK(flags);
				if (cnow.rng == cprev.rng && cnow.dsr == cprev.dsr &&
					cnow.dcd == cprev.dcd && cnow.cts == cprev.cts)
					return -EIO; /* no change => error */

				/* check_modem_status() handles ALTPIN mapping for the icount structures */
				if ( ((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
					 ((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
					 ((arg & TIOCM_CD)  && (cnow.dcd != cprev.dcd)) ||
					 ((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)) ) {
					return 0;
				}
				cprev = cnow;
			}
			/* NOTREACHED */

		/*
		 * Get counter of input serial line interrupts (DCD,RI,DSR,CTS)
		 * Return: write counters to the user passed counter struct
		 * NB: both 1->0 and 0->1 transitions are counted except for
		 *     RI where only 0->1 is counted.
		 */
		case TIOCGICOUNT:
			/* check_modem_status() handles ALTPIN mapping for the icount structures */
			DGLOCK(flags);
			cnow = info->state->icount;
			DGUNLOCK(flags);
			p_cuser = (struct serial_icounter_struct *) arg;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,1,0)
			error = putUser(cnow.cts, &p_cuser->cts);
			if (error) return error;
			error = putUser(cnow.dsr, &p_cuser->dsr);
			if (error) return error;
			error = putUser(cnow.rng, &p_cuser->rng);
			if (error) return error;
			error = putUser(cnow.dcd, &p_cuser->dcd);
			if (error) return error;
			error = putUser(cnow.rx, &p_cuser->rx);
			if (error) return error;
			error = putUser(cnow.tx, &p_cuser->tx);
			if (error) return error;
			error = putUser(cnow.frame, &p_cuser->frame);
			if (error) return error;
			error = putUser(cnow.overrun, &p_cuser->overrun);
			if (error) return error;
			error = putUser(cnow.parity, &p_cuser->parity);
			if (error) return error;
			error = putUser(cnow.brk, &p_cuser->brk);
			if (error) return error;
			error = putUser(cnow.buf_overrun, &p_cuser->buf_overrun);
			if (error) return error;			
#else
		putUser(cnow.cts, &p_cuser->cts);
		putUser(cnow.dsr, &p_cuser->dsr);
		putUser(cnow.rng, &p_cuser->rng);
		putUser(cnow.dcd, &p_cuser->dcd);
		putUser(cnow.rx, &p_cuser->rx);
		putUser(cnow.tx, &p_cuser->tx);
		putUser(cnow.frame, &p_cuser->frame);
		putUser(cnow.overrun, &p_cuser->overrun);
		putUser(cnow.parity, &p_cuser->parity);
		putUser(cnow.brk, &p_cuser->brk);
		putUser(cnow.buf_overrun, &p_cuser->buf_overrun);
#endif
			return 0;

		case TIOCSERGWILD:
		case TIOCSERSWILD:
			/* "setserial -W" is called in Debian boot */
			printk ("%s: TIOCSER?WILD ioctl obsolete, ignored.\n", DRVNAME);
			return 0;

		default:
			return -ENOIOCTLCMD;
		}
	return 0;
}

static void cpci_set_termios(struct tty_struct *tty, struct termios *old_termios)
{
	struct cpci_async_struct *info = (struct cpci_async_struct *)tty->driver_data;
	unsigned long flags;
	unsigned int cflag = tty->termios->c_cflag;
	
	if(tty->driver.subtype == SERIAL_TYPE_INFO)
		return;
	
	if (   (cflag == old_termios->c_cflag)
		&& (   RELEVANT_IFLAG(tty->termios->c_iflag)
		== RELEVANT_IFLAG(old_termios->c_iflag)))
	  return;

	change_speed(info, old_termios);

	/* Handle transition to B0 status */
	if ((old_termios->c_cflag & CBAUD) &&
		!(cflag & CBAUD)) {
		info->MCR &= ~(UART_MCR_DTR|UART_MCR_RTS);
		DGLOCK(flags);
		serial_out(info, UART_MCR, info->MCR);
		DGUNLOCK(flags);
	}
	
	/* Handle transition away from B0 status */
	if (!(old_termios->c_cflag & CBAUD) &&
		(cflag & CBAUD)) {
		info->MCR |= UART_MCR_DTR;
		if (!(tty->termios->c_cflag & CRTSCTS) ||
			!test_bit(TTY_THROTTLED, &tty->flags)) {
			info->MCR |= UART_MCR_RTS;
		}
		DGLOCK(flags);
		
		serial_out(info, UART_MCR, info->MCR);
		DGUNLOCK(flags);
	}
	
	/* Handle turning off CRTSCTS */
	if ((old_termios->c_cflag & CRTSCTS) &&
		!(tty->termios->c_cflag & CRTSCTS)) {
		tty->hw_stopped = 0;
		cpci_start(tty);
	}
}

/*
 * ------------------------------------------------------------
 * cpci_close()
 *
 * This routine is called when the serial port gets closed.  First, we
 * wait for the last remaining data to be sent.  Then, we unlink its
 * async structure from the interrupt chain if necessary, and we free
 * that IRQ if nothing is left in the chain.
 * ------------------------------------------------------------
 */
static void cpci_close(struct tty_struct *tty, struct file * filp)
{
	struct cpci_async_struct *info = (struct cpci_async_struct *)tty->driver_data;
	struct cpci_serial_state *state;
	unsigned long flags;

	if(tty->driver.subtype == SERIAL_TYPE_INFO) {
		MOD_DEC_USE_COUNT;
		return;
	}

	if (!info || serial_paranoia_check(info, tty->device, "cpci_close"))
		return;

	state = info->state;

	DGLOCK(flags);
	
	if (tty_hung_up_p(filp)) {
		DBG_CNT("before DEC-hung");
		MOD_DEC_USE_COUNT;
		DGUNLOCK(flags);
		return;
	}
	
	DEBUG_PRINTK("%s: cpci_close %s%d, count = %d\n", DRVNAME, parm_prefix, info->line, state->count);
	
	if ((tty->count == 1) && (state->count != 1)) {
		/*
		 * Uh, oh.  tty->count is 1, which means that the tty
		 * structure will be freed.  state->count should always
		 * be one in these conditions.  If it's greater than
		 * one, we've got real problems, since it means the
		 * serial port won't be shutdown.
		 */
		printk("%s: cpci_close: bad serial port count; tty->count is 1, "
			   "state->count is %d\n", DRVNAME, state->count);
		state->count = 1;
	}
	if (--state->count < 0) {
		printk("%s: cpci_close: bad serial port count for ttys%d: %d\n",
			   DRVNAME, info->line, state->count);
		state->count = 0;
	}
	if (state->count) {
		/* If the serial port is still open by another process, */
		/* this is where we realize it - and don't free up the  */
		/* other stuff - and return.				*/
		
		DBG_CNT("before DEC-2");
		MOD_DEC_USE_COUNT;
		DGUNLOCK(flags);
		//restore_flags(flags);
		return;
	}
	
	/* We do the rest of the stuff here - only if we're closing */
	/* off the entire device					*/
	
	info->flags |= ASYNC_CLOSING;

	if (info->flags & ASYNC_NORMAL_ACTIVE)
		info->state->normal_termios = *tty->termios;

	/*
	 * Now we wait for the transmit buffer to clear; and we notify
	 * the line discipline to only process XON/XOFF characters.
	 */
	tty->closing = 1;
	if (info->closing_wait != ASYNC_CLOSING_WAIT_NONE)
		tty_wait_until_sent(tty, info->closing_wait);
	/*
	 * At this point we stop accepting input.  To do this, we
	 * disable the receive line status interrupts, and tell the
	 * interrupt driver to stop checking the data ready bit in the
	 * line status register.
	 */
	info->IER &= ~UART_IER_RLSI;
	info->read_status_mask &= ~UART_LSR_DR;
	if (info->flags & ASYNC_INITIALIZED) {
		serial_out(info, UART_IER, info->IER);
		/*
		 * Before we drop DTR, make sure the UART transmitter
		 * has completely drained; this is especially
		 * important if there is a transmit FIFO!
		 */
		cpci_wait_until_sent(tty, info->timeout);
	}
	shutdown(info);
	if (tty->driver.flush_buffer)
		tty->driver.flush_buffer(tty);
	if (tty->ldisc.flush_buffer)
		tty->ldisc.flush_buffer(tty);
	tty->closing = 0;
	info->event = 0;
	info->tty = 0;
	if (info->blocked_open) {
		if (info->close_delay) {
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(info->close_delay);
		}
		wake_up_interruptible(&info->open_wait);
	}
	info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CLOSING);
	wake_up_interruptible(&info->close_wait);
	MOD_DEC_USE_COUNT;
	DGUNLOCK(flags);
	//restore_flags(flags);
}

/*
 * cpci_wait_until_sent() --- wait until the transmitter is empty
 */
static void cpci_wait_until_sent(struct tty_struct *tty, int timeout)
{
	struct cpci_async_struct * info = (struct cpci_async_struct *)tty->driver_data;
	unsigned long orig_jiffies, char_time;
	int lsr;
	
	if (serial_paranoia_check(info, tty->device, "cpci_wait_until_sent"))
		return;

	if (info->state->type == PORT_UNKNOWN)
		return;

	if (info->xmit_fifo_size == 0)
		return; /* Just in case.... */

	orig_jiffies = jiffies;
	/*
	 * Set the check interval to be 1/5 of the estimated time to
	 * send a single character, and make it at least 1.  The check
	 * interval should also be less than the timeout.
	 *
	 * Note: we have to use pretty tight timings here to satisfy
	 * the NIST-PCTS.
	 */
	char_time = (info->timeout - HZ/50) / info->xmit_fifo_size;
	char_time = char_time / 5;
	if (char_time == 0)
		char_time = 1;
	if (timeout)
	  char_time = MIN(char_time, timeout);
	/*
	 * If the transmitter hasn't cleared in twice the approximate
	 * amount of time to send the entire FIFO, it probably won't
	 * ever clear.  This assumes the UART isn't doing flow
	 * control, which is currently the case.  Hence, if it ever
	 * takes longer than info->timeout, this is probably due to a
	 * UART bug of some kind.  So, we clamp the timeout parameter at
	 * 2*info->timeout.
	 */
	if (!timeout || timeout > 2*info->timeout)
		timeout = 2*info->timeout;
	
	DEBUG_PRINTK("%s: In cpci_wait_until_sent(%d) check=%lu...", DRVNAME, timeout, char_time);
	DEBUG_PRINTK("%s: jiff=%lu...", DRVNAME, jiffies);
	
	while (!((lsr = serial_in(info, UART_LSR)) & UART_LSR_TEMT)) {
		DEBUG_PRINTK("%s: lsr = %d (jiff=%lu)...", DRVNAME, lsr, jiffies);
		
		current->state = TASK_INTERRUPTIBLE;
//        current->counter = 0;	/* make us low-priority */
		schedule_timeout(char_time);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)	
		if (current->signal & ~current->blocked)
			break;
#else
		if (signal_pending(current))
			break;
#endif

		if (timeout && time_after(jiffies, orig_jiffies + timeout))
			break;
	}
	current->state = TASK_RUNNING;
	DEBUG_PRINTK("%s: lsr = %d (jiff=%lu)...done\n", DRVNAME, lsr, jiffies);
}

/*
 * cpci_hangup() --- called by tty_hangup() when a hangup is signaled.
 */
static void cpci_hangup(struct tty_struct *tty)
{
	struct cpci_async_struct * info = (struct cpci_async_struct *)tty->driver_data;
	struct cpci_serial_state *state = info->state;
	
	if(tty->driver.subtype == SERIAL_TYPE_INFO)
		return;
		
	if (serial_paranoia_check(info, tty->device, "cpci_hangup"))
		return;

	state = info->state;
	
	cpci_flush_buffer(tty);
	shutdown(info);
	info->event = 0;
	state->count = 0;
	info->flags &= ~(ASYNC_NORMAL_ACTIVE);
	info->tty = 0;
	wake_up_interruptible(&info->open_wait);
}

/*
 * ------------------------------------------------------------
 * cpci_open() and friends
 * ------------------------------------------------------------
 */
static int block_til_ready(struct tty_struct *tty, struct file * filp,
			   struct cpci_async_struct *info)
{
	
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
	DECLARE_WAITQUEUE(wait, current);
#else
	struct wait_queue wait = { current, NULL };
#endif

	struct cpci_serial_state *state = info->state;
	int		retval;
	int		do_clocal = 0, extra_count = 0;
	unsigned long	flags;

	/*
	 * If the device is in the middle of being closed, then block
	 * until it's done, and then try again.
	 */
	if (tty_hung_up_p(filp) ||
		(info->flags & ASYNC_CLOSING)) {
		if (info->flags & ASYNC_CLOSING)
			interruptible_sleep_on(&info->close_wait);
#ifdef SERIAL_DO_RESTART
		return ((info->flags & ASYNC_HUP_NOTIFY) ?
			-EAGAIN : -ERESTARTSYS);
#else
		return -EAGAIN;
#endif
	}

	/*
	 * If non-blocking mode is set, or the port is not enabled,
	 * then make the check up front and then exit.
	 */
	if ((filp->f_flags & O_NONBLOCK) ||
		(tty->flags & (1 << TTY_IO_ERROR))) {
		info->flags |= ASYNC_NORMAL_ACTIVE;
		return 0;
	}

	if (tty->termios->c_cflag & CLOCAL)
		do_clocal = 1;
	
	/*
	 * Block waiting for the carrier detect and the line to become
	 * free (i.e., not in use by the callout).  While we are in
	 * this loop, state->count is dropped by one, so that
	 * cpci_close() knows when to free things.  We restore it upon
	 * exit, either normal or abnormal.
	 */
	retval = 0;
	add_wait_queue(&info->open_wait, &wait);
	DEBUG_PRINTK("%s: block_til_ready before block: ttys%d, count = %d\n",
		   DRVNAME, state->line, state->count);
	DGLOCK(flags);
	if (!tty_hung_up_p(filp)) {
		extra_count = 1;
		state->count--;
	}
	DGUNLOCK(flags);
	info->blocked_open++;
	while (1) {
		DGLOCK(flags);
		if ((tty->termios->c_cflag & CBAUD))
			info->MCR |= (UART_MCR_DTR | UART_MCR_RTS);
			serial_out(info, UART_MCR, info->MCR);
			
			DGUNLOCK(flags);
		current->state = TASK_INTERRUPTIBLE;
		if (tty_hung_up_p(filp) ||
			!(info->flags & ASYNC_INITIALIZED)) {
#ifdef SERIAL_DO_RESTART
			if (info->flags & ASYNC_HUP_NOTIFY)
				retval = -EAGAIN;
			else	{
				retval = -ERESTARTSYS;	
				}
#else
			retval = -EAGAIN;
#endif
			break;
		}
		if (!(info->flags & ASYNC_CLOSING) &&
			(do_clocal || (serial_in(info, UART_MSR) &
				   (info->digiext.digi_flags & DIGI_ALTPIN?
				   UART_MSR_DSR : UART_MSR_DCD))))
			break;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
				if (current->signal & ~current->blocked)
#else
		if(signal_pending(current))
#endif
		{
					retval = -ERESTARTSYS;
			break;
		}

		DEBUG_PRINTK("%s: block_til_ready blocking: ttys%d, count = %d\n",
			   DRVNAME, info->line, state->count);
		schedule();
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&info->open_wait, &wait);
	if (extra_count)
		state->count++;
	info->blocked_open--;
	DEBUG_PRINTK("%s: block_til_ready after blocking: ttys%d, count = %d\n",
		   DRVNAME, info->line, state->count);
	
	if (retval)
		return retval;
	info->flags |= ASYNC_NORMAL_ACTIVE;
	
	return 0;
}

static int get_cpci_async_struct(int line, struct cpci_async_struct **ret_info)
{
	struct cpci_async_struct *info;
	struct cpci_serial_state *sstate;

	sstate = cpci_table + line;
	sstate->count++;
	if (sstate->info) {
		*ret_info = sstate->info;
		return 0;
	}
	info = kmalloc(sizeof(struct cpci_async_struct), GFP_KERNEL);
	if (!info) {
		sstate->count--;
		return -ENOMEM;
	}
	memset(info, 0, sizeof(struct cpci_async_struct));
	init_waitqueue_head(&info->open_wait);
	init_waitqueue_head(&info->close_wait);
	init_waitqueue_head(&info->delta_msr_wait);
	info->magic = CPCI_MAGIC;
	info->port = sstate->port;
	info->flags = sstate->flags;
	info->io_type = sstate->io_type;
	info->iomem_base = sstate->iomem_base;
	info->io_flags = sstate->io_flags;
	info->xmit_fifo_size = sstate->xmit_fifo_size;
	info->line = line;
	info->tqueue.routine = do_softint;
	info->tqueue.data = info;
	info->state = sstate;
	/* Set trigger levels here (on the first open) so subsequent user changes remain sticky */
	if( info->state->type == PORT_17158) {
		info->digiext.rxtrg=TRIGGER_RX_17158;
		info->digiext.txtrg = TRIGGER_TX_17158;
	} else {
		/* Not applicable */
		info->digiext.rxtrg = info->digiext.txtrg = 0;
	}

	if (sstate->info) {
		kfree( info );
		*ret_info = sstate->info;
		return 0;
	}
	*ret_info = sstate->info = info;
	return 0;
}

/*
 * This routine is called whenever a serial port is opened.  It
 * enables interrupts for a serial port, linking in its async structure into
 * the IRQ chain.   It also performs the serial-specific
 * initialization for the tty structure.
 */
static int cpci_open(struct tty_struct *tty, struct file * filp)
{
	struct cpci_async_struct	*info;
	int 			retval, line;
	unsigned long		page;

	MOD_INC_USE_COUNT;
	
	/* if this is the management minor, just return. there's nothing */
	/* we should be doing here, other than that. */
	
	if(tty->driver.subtype == SERIAL_TYPE_INFO)
		return(0);
	
	line = MINOR(tty->device) - tty->driver.minor_start;
	if ((line < 0) || (line >= NR_PORTS)) {
		MOD_DEC_USE_COUNT;
		return -ENODEV;
	}
	retval = get_cpci_async_struct(line, &info);
	if (retval) {
		printk("%s: debug: cpci_open() get_cpci_async_struct() failed!\n", DRVNAME);
		MOD_DEC_USE_COUNT;
		return retval;
	}
	
	/*info->loopedback = 1;*/
	info->loopedback = 0;	/* Turn off internal UART loopback mode */
	
	tty->driver_data = info;
	info->tty = tty;
	if (serial_paranoia_check(info, tty->device, "cpci_open")) {
		/* MOD_DEC_USE_COUNT; "info->tty" will cause this */
		return -ENODEV;
	}

	DEBUG_PRINTK("%s: cpci_open %s%d, count = %d\n", DRVNAME, tty->driver.name, info->line,
		   info->state->count);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	info->tty->low_latency = (info->flags & ASYNC_LOW_LATENCY) ? 1 : 0;
#endif

	if (!tmp_buf) {
		page = get_free_page(GFP_KERNEL);
		if (!page) {
			/* MOD_DEC_USE_COUNT; "info->tty" will cause this? */
			return -ENOMEM;
		}
		if (tmp_buf)
			free_page(page);
		else
			tmp_buf = (unsigned char *) page;
	}

	/*
	 * If the port is the middle of closing, bail out now
	 */
	if (tty_hung_up_p(filp) ||
		(info->flags & ASYNC_CLOSING)) {
		if (info->flags & ASYNC_CLOSING) {
			interruptible_sleep_on(&info->close_wait);
			/* MOD_DEC_USE_COUNT; "info->tty" will cause this? */
#ifdef SERIAL_DO_RESTART
			return ((info->flags & ASYNC_HUP_NOTIFY) ?
				-EAGAIN : -ERESTARTSYS);
#else
			return -EAGAIN;
#endif
		}
	
	}

	 /*
	 * Start up serial port
	 */
	retval = startup(info);
	if (retval) {
		/* MOD_DEC_USE_COUNT; "info->tty" will cause this? */
		printk("%s: debug: cpci_open() - startup() call failed!\n", DRVNAME);
		return retval;
	}

	if(info->state->count == 1 && !info->blocked_open) {
		retval = block_til_ready(tty, filp, info);
		if (retval) {
			/* MOD_DEC_USE_COUNT; "info->tty" will cause this? */
			DEBUG_PRINTK("%s: cpci_open returning after block_til_ready with %d\n",
				   DRVNAME, retval);
			return retval;
		}
	}
	
	if ((info->state->count == 1) &&
		(info->flags & ASYNC_SPLIT_TERMIOS)) {
		if (tty->driver.subtype == SERIAL_TYPE_NORMAL)
			*tty->termios = info->state->normal_termios;
		
		change_speed(info, 0);
	}

	info->session = current->session;
	info->pgrp = current->pgrp;
	
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	tty->low_latency = 1;
#endif
	
	DEBUG_PRINTK("%s: cpci_open ttys%d successful...\n", DRVNAME, info->line);
	return 0;
}

/*
 * ---------------------------------------------------------------------
 * cpci_init() and friends
 *
 * cpci_init() is called at boot-time to initialize the serial driver.
 * ---------------------------------------------------------------------
 */

/*
 * This is a quickie test to see how big the FIFO is.
 * It doesn't work at all the time, more's the pity.
 */

static int size_fifo(struct cpci_async_struct *info)
{
	unsigned char old_fcr, old_mcr, old_dll, old_dlm;
	int count;

	old_fcr = serial_in(info, UART_FCR);
	old_mcr = serial_in(info, UART_MCR);
	serial_out(info, UART_FCR, UART_FCR_ENABLE_FIFO |
	UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT);
	serial_out(info, UART_MCR, UART_MCR_LOOP);
	serial_out(info, UART_LCR, UART_LCR_DLAB);
	old_dll = serial_in(info, UART_DLL);
	old_dlm = serial_in(info, UART_DLM);
	serial_out(info, UART_DLL, 0x01);
	serial_out(info, UART_DLM, 0x00);
	serial_out(info, UART_LCR, 0x03);
	for (count = 0; count < 256; count++)
	serial_out(info, UART_TX, count);
	mdelay(20);
	for (count = 0; (serial_in(info, UART_LSR) & UART_LSR_DR) &&
	(count < 256); count++)
	serial_in(info, UART_RX);
	serial_out(info, UART_FCR, old_fcr);
	serial_out(info, UART_MCR, old_mcr);
	serial_out(info, UART_LCR, UART_LCR_DLAB);
	serial_out(info, UART_DLL, old_dll);
	serial_out(info, UART_DLM, old_dlm);
	return count;
}

/*
 * This is a helper routine to autodetect StarTech/Exar/Oxsemi UART's.
 * When this function is called we know it is at least a StarTech
 * 16650 V2, but it might be one of several StarTech UARTs, or one of
 * its clones.  (We treat the broken original StarTech 16650 V1 as a
 * 16550, and why not?  Startech doesn't seem to even acknowledge its
 * existence.)
 *
 * What evil have men's minds wrought...
 */

static void autoconfig_startech_uarts(struct cpci_async_struct *info,
					  struct cpci_serial_state *state,
					  unsigned long flags)
{
	unsigned char scratch, scratch2, scratch3;
	
	/*
	 * First we check to see if it's an Oxford Semiconductor UART.
	 *
	 * If we have to do this here because some non-National
	 * Semiconductor clone chips DGLOCK up if you try writing to the
	 * LSR register (which serial_icr_read does)
	 */
	if (state->type == PORT_16550A) {
		/* Check for Oxford Semiconductor 16C950 */
		scratch = serial_icr_read(info, UART_ID1);
		scratch2 = serial_icr_read(info, UART_ID2);
		scratch3 = serial_icr_read(info, UART_ID3);
		
		if (scratch == 0x16 && scratch2 == 0xC9 &&
			(scratch3 == 0x50 || scratch3 == 0x52 ||
			 scratch3 == 0x54)) {
			state->type = PORT_16C950;
			state->revision = serial_icr_read(info, UART_REV) |
				(scratch3 << 8);
			return;
		}
	}
	
	/*
	 * We check for a XR16C850 by setting DLL and DLM to 0, and
	 * then reading back DLL and DLM.  If DLM reads back 0x10,
	 * then the UART is a XR16C850 and the DLL contains the chip
	 * revision.  If DLM reads back 0x14, then the UART is a
	 * XR16C854.
	 *
	 */
	serial_out(info, UART_LCR, UART_LCR_DLAB);
	serial_out(info, UART_DLL, 0);
	serial_out(info, UART_DLM, 0);
	state->revision = serial_in(info, UART_DLL);
	scratch = serial_in(info, UART_DLM);
	serial_out(info, UART_LCR, 0);
	if (scratch == 0x10 || scratch == 0x14) {
		state->type = PORT_16850;
		return;
	}

	/*
	 * We distinguish between the '654 and the '650 by counting
	 * how many bytes are in the FIFO.  I'm using this for now,
	 * since that's the technique that was sent to me in the
	 * serial driver update, but I'm not convinced this works.
	 * I've had problems doing this in the past.  -TYT
	 */
	if (size_fifo(info) == 64) {
		state->type = PORT_16654;
	} else {
		state->type = PORT_16650V2;
	}
}



/*
 * This routine is called by cpci_init() to initialize a specific serial
 * port.  It determines what type of UART chip this serial port is
 * using: 8250, 16450, 16550, 16550A.  The important question is
 * whether or not this UART is a 16550A, since this will determine
 * whether or not we can use its FIFO features.
 */


static void autoconfig(struct cpci_serial_state * state)
{
	unsigned char status1, status2, scratch, scratch2;
	struct cpci_async_struct *info, scr_info;
	unsigned long flags;
	int EFR;
	
	state->type = PORT_UNKNOWN;
	info = &scr_info;	/* This is just for serial_{in,out} */

	info->magic = CPCI_MAGIC;
	info->port = state->port;
	info->flags = state->flags;

	info->io_type = state->io_type;
	info->iomem_base = state->iomem_base;
	info->io_flags = state->io_flags;
	
	if(!CONFIGURED_SERIAL_PORT(info))
		return;
	
	DGLOCK(flags);

	/*
	 * Do a simple existence test first; if we fail this, there's
	 * no point trying anything else.
	 *
	 * 0x80 is used as a nonsense port to prevent against false
	 * positives due to ISA bus float.  The assumption is that
	 * 0x80 is a non-existent port; which should be safe since
	 * include/asm/io.h also makes this assumption.
	 */
	scratch = serial_in(info, UART_IER);
	serial_out(info, UART_IER, 0);
	outb(0xff, 0x080);
	scratch2 = serial_in(info, UART_IER);
	serial_out(info, UART_IER, scratch);
	if (scratch2) {
		DGUNLOCK(flags);
		//restore_flags(flags);
		return;		/* We failed; there's nothing here */
	}
	/*
	 * Check to see if a UART is really there.  Certain broken
	 * internal modems based on the Rockwell chipset fail this
	 * test, because they apparently don't implement the loopback
	 * test mode.  So this test is skipped on the COM 1 through
	 * COM 4 ports.  This *should* be safe, since no board
	 * manufacturer would be stupid enough to design a board
	 * that conflicts with COM 1-4 --- we hope!
	 */
	if (!(state->flags & ASYNC_SKIP_TEST)) {
		scratch = serial_in(info, UART_MCR);
		serial_out(info, UART_MCR, UART_MCR_LOOP | scratch);
		serial_out(info, UART_MCR, UART_MCR_LOOP | 0x0A);
		status1 = serial_in(info, UART_MSR) & 0xF0;
		serial_out(info, UART_MCR, scratch);
		if (status1 != 0x90) {
			DGUNLOCK(flags);
			//restore_flags(flags);
			return;
		}
	}
	
	scratch2 = serial_in(info, UART_LCR);
	
	if(state->io_flags & CPCI_IOFLAGS_EXTENDED_EXTREGS) {
		EFR = UART_EXTENDED_EFR;
		serial_out(info, EFR, 0);
	} else {
		EFR = UART_EFR;
		serial_out(info, UART_LCR, 0xBF); /* set up for StarTech test */
		serial_out(info, EFR, 0);	/* EFR is the same as FCR */
		serial_out(info, UART_LCR, 0);
	}
	
	serial_out(info, UART_FCR, UART_FCR_ENABLE_FIFO);
	scratch = serial_in(info, UART_ISR) >> 6;
	switch (scratch) {
		case 0:
			state->type = PORT_16450;
			break;
		case 1:
			state->type = PORT_UNKNOWN;
			break;
		case 2:
			state->type = PORT_16550;
			break;
		case 3:
			state->type = PORT_16550A;
			break;
	}
	
	if (state->type == PORT_16550A) {
		/* Check for Startech UART's */
		
		/* make sure that we write 0xBF in places we need.. */
		
		serial_out(info, UART_LCR, UART_LCR_DLAB);
		if (serial_in(info, UART_EFR) == 0) {
			state->type = PORT_16650;
		} else {
		
		
			/* If it supports extended-mapped registers, it most      */
			/* certainly is NOT a 16650. And since there is currently */
			/* only 1 card that uses this flag, we'll call it the 	  */
			/* 17C158.						  */
				
			if(info->io_flags & CPCI_IOFLAGS_EXTENDED_EXTREGS) {
				state->type = PORT_17158;
			} else {
				serial_out(info, UART_LCR, 0xBF);
				if(serial_in(info, UART_EFR) == 0)
					autoconfig_startech_uarts(info, state, flags);
			}
		}			
	}
	
	if (state->type == PORT_16550A) {
		/* Check for TI 16750 */
		serial_out(info, UART_LCR, scratch2 | UART_LCR_DLAB);
		serial_out(info, UART_FCR,
				UART_FCR_ENABLE_FIFO | UART_FCR7_64BYTE);
		scratch = serial_in(info, UART_ISR) >> 5;
		if (scratch == 7) {
			serial_out(info, UART_LCR, 0);
			scratch = serial_in(info, UART_ISR) >> 5;
			if (scratch == 6) {
				state->type = PORT_16750;
			}
		}
		serial_out(info, UART_FCR, UART_FCR_ENABLE_FIFO);
	}
	serial_out(info, UART_LCR, scratch2);
	if (state->type == PORT_16450) {
		scratch = serial_in(info, UART_SCR);
		serial_out(info, UART_SCR, 0xa5);
		status1 = serial_in(info, UART_SCR);
		serial_out(info, UART_SCR, 0x5a);
		status2 = serial_in(info, UART_SCR);
		serial_out(info, UART_SCR, scratch);

		if ((status1 != 0xa5) || (status2 != 0x5a)) {
			state->type = PORT_8250;
		}
	}
	state->xmit_fifo_size =	uart_config[state->type].dfl_xmit_fifo_size;

	if (state->type == PORT_UNKNOWN) {
		DGUNLOCK(flags);
		//restore_flags(flags);
		return;
	}

	 /* Only request it if we've got a set of I/O ports. MMAP don't count. CWS */

	if(state->port)
		request_region(info->port,8,"serial(auto)");

	/*
	 * Reset the UART.
	 */
	serial_out(info, UART_MCR, 0x00);
	serial_out(info, UART_FCR, (UART_FCR_CLEAR_RCVR |
					 UART_FCR_CLEAR_XMIT));
	(void)serial_in(info, UART_RX);
	serial_out(info, UART_IER, 0);
	
	DGUNLOCK(flags);
	//restore_flags(flags);
}

/*
 * The serial driver boot-time initialization code!
 */

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
int __init cpci_init(void)
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
int cpci_init(void)
#else
__initfunc(int cpci_init(void))
#endif

{
	int i;
	int retval;
	struct cpci_serial_state * state;
	extern void atomwide_serial_init (void);
	extern void dualsp_serial_init (void);

#ifdef CONFIG_ATOMWIDE_SERIAL
	atomwide_serial_init ();
#endif
#ifdef CONFIG_DUALSP_SERIAL
	dualsp_serial_init ();
#endif
	
	serial_clear_cpci_table();
	if(init_boards() == -EBUSY)
		/*return(-EBUSY);*/
		return(0);
		
	if(board[0].membase) {
		cpci_internal_init();
	}

	/* - REMOVED - init_bh(SERIAL_BH, do_serial_bh);*/
	
	init_timer( &cpci_timer_struct );
	cpci_timer_struct.function = cpci_timer;

	for (i = 0; i < NR_IRQS; i++) {
		IRQ_ports[i] = 0;
		IRQ_timeout[i] = 0;
	}

	/* Initialize the tty_driver structure */
	
	memset(&serial_driver, 0, sizeof(struct tty_driver));
	serial_driver.magic = TTY_DRIVER_MAGIC;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	serial_driver.driver_name = "ClassicPCI";
#endif
	/* Leave as 'ttyCP" so mknod scripts don't have to know the parm_prefix */
	serial_driver.name = "ttyCP";
	/*serial_driver.name = parm_prefix;*/
	/*serial_driver.name = "ClassicPCI";*/
	serial_driver.major = 0;
	serial_driver.minor_start = 0;
	serial_driver.num = NR_PORTS;
	serial_driver.type = TTY_DRIVER_TYPE_SERIAL;
	serial_driver.subtype = SERIAL_TYPE_NORMAL;
	serial_driver.init_termios = tty_std_termios;
	serial_driver.init_termios.c_cflag =
		B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	serial_driver.flags = TTY_DRIVER_REAL_RAW;
	serial_driver.refcount = &serial_refcount;
	serial_driver.table = serial_table;
	serial_driver.termios = serial_termios;
	serial_driver.termios_locked = serial_termios_locked;

	serial_driver.open = cpci_open;
	serial_driver.close = cpci_close;
	serial_driver.write = cpci_write;
	serial_driver.put_char = cpci_put_char;
	serial_driver.flush_chars = cpci_flush_chars;
	serial_driver.write_room = cpci_write_room;
	serial_driver.chars_in_buffer = cpci_chars_in_buffer;
	serial_driver.flush_buffer = cpci_flush_buffer;
	serial_driver.ioctl = cpci_ioctl;
	serial_driver.throttle = cpci_throttle;
	serial_driver.unthrottle = cpci_unthrottle;
	serial_driver.set_termios = cpci_set_termios;
	serial_driver.stop = cpci_stop;
	serial_driver.start = cpci_start;
	serial_driver.hangup = cpci_hangup;
	
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,1,0)
	serial_driver.send_xchar = cpci_send_xchar;
	serial_driver.break_ctl = cpci_break;
	serial_driver.wait_until_sent = cpci_wait_until_sent;
#endif
	
#ifndef OLDPROC
	serial_driver.read_proc = cpci_read_proc;
#endif //OLDPROC
	
	mgmt_driver = serial_driver;
	/*mgmt_driver.name = "ClassicMGMT";*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	mgmt_driver.driver_name = "ClassicMGMT";
#endif
	mgmt_driver.name = "cpciCtl";
	mgmt_driver.init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL;
	mgmt_driver.subtype = SERIAL_TYPE_INFO;
	mgmt_driver.table = mgmt_table;
	mgmt_driver.termios = mgmt_termios;
	mgmt_driver.termios_locked = mgmt_termios_locked;
	mgmt_driver.num = 1;

	retval = tty_register_driver(&serial_driver);
	if(retval < 0) panic("Couldn't register serial driver\n");
	
	/* Register management device */
	
	retval = tty_register_driver(&mgmt_driver);
	if(retval < 0) panic("Couldn't register mgmt_driver\n");
	
	DEBUG_PRINTK("%d ports supported\n",NR_PORTS);
	
	for (i = 0, state = cpci_table; i < NR_PORTS; i++,state++) {
		state->magic = CPCI_SSTATE_MAGIC;
		state->line = i;
		state->type = PORT_UNKNOWN;
		state->custom_divisor = 0;
		state->close_delay = 5*HZ/10;
		state->closing_wait = 30*HZ;
		state->normal_termios = serial_driver.init_termios;
		state->icount.cts = state->icount.dsr =
			state->icount.rng = state->icount.dcd = 0;
		state->icount.rx = state->icount.tx = 0;
		state->icount.frame = state->icount.parity = 0;
		state->icount.overrun = state->icount.brk = 0;
		state->irq = irq_cannonicalize(state->irq);
		/* If we've got a port to work with, use this. CWS */
		if (state->port && check_region(state->port,8)) {
			continue;
		}
		if (state->flags & ASYNC_BOOT_AUTOCONF) {
			autoconfig(state);
		}

	}
	/*
	 * Detect the IRQ only once every port is initialised,
	 * because some 16450 do not reset to 0 the MCR register.
	 */
	for (i = 0, state = cpci_table; i < NR_PORTS; i++,state++) {
		if (state->type == PORT_UNKNOWN)
			continue;
	
		printk(KERN_INFO "%s%d at 0x%x (irq = %d) is a %s\n",
		parm_prefix, state->line+parm_offset,
		(state->iomem_base) ? (unsigned int)state->iomem_base : state->port,
		state->irq,
		uart_config[state->type].name);
	 	
	}
	return(0);
}


#ifdef MODULE
int init_module(void)
{
	int ret;
	ret=cpci_init();
#ifdef OLDPROC
	proc_register_dynamic (&proc_root, &cpci_proc_entry);
#endif
	return (ret);
}

void cleanup_module(void)
{
	unsigned long flags;
	int i;
	int e1;
	int e2;
	
	struct cpci_async_struct *info;

	if(!cpci_totalboards)
		return;

	cpci_internal_cleanup();
	
	DGLOCK(flags);
	del_timer_sync( &cpci_timer_struct );
	if ((e1 = tty_unregister_driver(&serial_driver)))
		printk("%s: failed to unregister serial driver (%d)\n",
			DRVNAME, e1);
	if ((e2 = tty_unregister_driver(&mgmt_driver)))
		printk("%s: failed to unregister managemnt driver (%d)\n",
			DRVNAME, e2);	

	DGUNLOCK(flags);

	for (i = 0; i < NR_PORTS; i++) {
		if (cpci_table[i].type != PORT_UNKNOWN) {
			/* If operating in (the now default) memory-mapped mode,
			 * port will be 0 and no region will have been requested.
			 * Only release region if operating in I/O-mapped mode.
			 */
			if( cpci_table[i].port )
				release_region(cpci_table[i].port, 8);
		}
		info = cpci_table[i].info;
		if (info) {
			cpci_table[i].info = NULL;
			kfree( info );
		}
	}
	
	if (tmp_buf) {
		free_page((unsigned long) tmp_buf);
		tmp_buf = NULL;
	}
#ifndef OLDPCI
	pci_unregister_driver (&cpci_driver);
#endif
#ifdef OLDPROC
	proc_unregister (&proc_root, cpci_proc_entry.low_ino);
#endif
}
#endif /* MODULE */
