/*
 * Private header file for the ClassicPCI serial driver
 * Originally based upon serialP.h, for the dumb serial driver in Linux.
 * Ted's credits are below
 *
 * Private header file for the (dumb) serial driver
 *
 * Copyright (C) 1997 by Theodore Ts'o.
 * 
 * Redistribution of this file is permitted under the terms of the GNU 
 * Public License (GPL)
 */

/*
 * This is our internal structure for each serial port's state.
 * 
 * Many fields are paralleled by the structure used by the serial_struct
 * structure.
 *
 * For definitions of the flags field, see tty.h
 */

#ifdef KERNEL
#include <linux/termios.h>
#endif
#include <linux/tqueue.h>


/*
 * Defined macros for ClassicPCI
 */
 

/*
 * Counters of the input lines (CTS, DSR, RI, CD) interrupts
 */
struct cpci_async_icount {
	__u32	cts, dsr, rng, dcd, tx, rx;
	__u32	frame, parity, overrun, brk;
	__u32	buf_overrun;
};

struct cpci_serial_state {
	int	magic;
	int	baud_base;
	int	port;
	int	irq;
	int	flags;
	int	hub6;
	int	type;
	int	line;
	int 	revision;	/* 950 */
	int	xmit_fifo_size;
	int	custom_divisor;
	int	count;
	unsigned char    *iomem_base;
	char    io_type;
	unsigned short io_flags;
	unsigned short	close_delay;
	unsigned short	closing_wait; /* time to wait before closing */
	struct cpci_async_icount	icount;	
	struct termios		normal_termios;
	struct termios		callout_termios;
	struct cpci_async_struct *info;
};

struct cpci_async_struct {
	int			magic;
	int			port;
	int			hub6;
	int			flags;
	int			xmit_fifo_size;
	struct cpci_serial_state	*state;
	struct tty_struct 	*tty;
	int			read_status_mask;
	int			ignore_status_mask;
	int			timeout;
	int			quot;
	int			x_char;	/* xon/xoff character */
	int			close_delay;
	unsigned short		closing_wait;
	unsigned short		closing_wait2;
        int                     IER;    /* Interrupt Enable Register */
        int                     MCR;    /* Modem control register */
        int                     LCR;    /* Line control register */
        int                     ACR;     /* 16950 Additional Control Reg. */
	unsigned long		event;
	unsigned long		last_active;
	int			line;
	int			blocked_open; /* # of blocked opens */
	long			session; /* Session of opening process */
	long			pgrp; /* pgrp of opening process */
	unsigned char 		*xmit_buf;
	int			xmit_head;
	int			xmit_tail;
	int			xmit_cnt;
	char			*iomem_base;
	char			io_type;
	unsigned short		io_flags;
	struct tq_struct	tqueue;
	unsigned char		ch_pscan_state;
	unsigned char		ch_pscan_savechar;
	char			throttled;	
#if LINUX_VERSION_CODE > 0x020300
	wait_queue_head_t	open_wait;
	wait_queue_head_t	close_wait;
	wait_queue_head_t	delta_msr_wait;
#else	
	struct wait_queue	*open_wait;
	struct wait_queue	*close_wait;
	struct wait_queue	*delta_msr_wait;
#endif

	char			loopedback;
	struct digi_struct digiext;	/* Digi-specific extra info */
	
	struct cpci_async_struct	*next_port; /* For the linked list */
	struct cpci_async_struct	*prev_port;


	char myflipbuf[MYFLIPLEN];
};

#define CPCI_MAGIC 0x5301
#define CPCI_SSTATE_MAGIC 0x5302

/*
 * The size of the serial xmit buffer is 1 page, or 4096 bytes
 */
#define CPCI_XMIT_SIZE 4096

/*
 * Events are used to schedule things to happen at timer-interrupt
 * time, instead of at rs interrupt time.
 */
#define CPCI_EVENT_WRITE_WAKEUP	0

